unit GabeImage;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TOnMouseEnter = Procedure(Sender: TObject) of Object;
  TOnMouseLeave = Procedure(Sender: TObject) of Object;

  TGabeImage = class(TGraphicControl)
  private
    FBitmap: TBitmap;
    FAutoSize: Boolean;
    FOnMouseEnter: TOnMouseEnter;
    FOnMouseLeave: TOnMouseLeave;
    FCursDnHandle: Integer;
    FCursUpHandle: Integer;
    FCursDefHandle: Integer;
    FCursDisHandle: Integer;
    FTransparent: Boolean;
    FCursOn: Boolean;
    Destructor Destroy; override;
    procedure SetImage(const Value: TBitmap);
    procedure Paint; override;
    procedure SetAutoSize(const Value: Boolean);
    procedure CMMouseEnter(var msg:TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
    procedure SetCursHandle(Index: Integer;const Value: Integer);
    function  GetCursHandle(Index: Integer): Integer;
    procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MouseUp(Sender: TObject; Button: TMouseButton;  Shift: TShiftState; X, Y: Integer);
    procedure SetTransparent(const Value: Boolean);
    procedure SetCursOn(const Value: Boolean);
  protected
  public
    Constructor Create(AOwner: TComponent); override;
  published
     property Image: TBitmap read FBitmap write SetImage;
     property AutoSize: Boolean read FAutoSize write SetAutoSize;
     property OnMouseEnter: TOnMouseEnter read FOnMouseEnter write FOnMouseEnter;
     property OnMouseLeave: TOnMouseLeave read FOnMouseLeave write FOnMouseLeave;
     property Transparent: Boolean read FTransparent write SetTransparent;

     {Set cursor properties}

     property CursDnHandle: Integer index 0 read GetCursHandle write SetCursHandle;
     property CursUpHandle: Integer index 1 read GetCursHandle write SetCursHandle;
     property CursDefHandle: Integer index 2 read GetCursHandle write SetCursHandle;
     property CursDisHandle: Integer index 3 read GetCursHandle write SetCursHandle;
     property Curson: boolean read FCurson write SetCurson default True;
     {Standard Events & Properties}
     property Align;
     property DragKind;
     property DragCursor;
     property DragMode;
     property Enabled;
     property Hint;
     property PopupMenu;
     property Visible;
     property Color;
     property ParentColor;
     property OnClick;
     property OnDblClick;
     property OnDragDrop;
     property OnDragOver;
     property OnEndDock;
     property OnEndDrag;
     property OnMouseDown;
     property OnMouseMove;
     property OnMouseUp;
     property OnStartDock;
     property OnStartDrag;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Gabes', [TGabeImage]);
end;

{ TNewImage }


constructor TGabeImage.Create(AOwner: TComponent);
begin
   Inherited Create(AOwner);
   FBitmap := TBitmap.Create;
   Width := 32;
   Height := 32;
   OnMouseDown := MouseDown;
   OnMouseUp := MouseUp;
   FCurson:=true;
end;

destructor TGabeImage.Destroy;
begin
   FBitmap.Free;
   Inherited Destroy;
end;

procedure TGabeImage.Paint;
var Dest:TRect;
begin
   Inherited;
   if csDesigning in ComponentState then
      with inherited Canvas do begin
         Pen.Style := psDash;
         Brush.Style := bsClear;
         Rectangle(0,0,Width,Height);
       end;

   If (FBitMap.Width = 0) or (FBitmap.Height = 0) then
      With Inherited Canvas do begin
         Canvas.Brush.Style := bsClear;
         Canvas.Brush.Color := clNone;
         Exit;
      End;

   If FTransparent then begin
      FBitmap.Transparent := True;
      FBitmap.TransparentColor := FBitmap.Canvas.Pixels[0, FBitmap.Height];
      FBitmap.TransparentMode := tmAuto;
   End Else begin
      FBitmap.Transparent := False;
      FBitmap.TransparentColor := clNone; //FBitmap.Canvas.Pixels[0, FBitmap.Height];
      FBitmap.TransparentMode := tmAuto;
   End;

   Canvas.Draw(0, 0, FBitmap);
   If AutoSize then begin
      Width := FBitmap.Width;
      Height := FBitmap.Height;
   End;

end;

procedure TGabeImage.SetAutoSize(const Value: Boolean);
begin
  FAutoSize := Value;
  Paint;
end;

procedure TGabeImage.SetImage(const Value: TBitmap);
begin
  FBitmap.Assign(Value);
  Paint;
end;

procedure TGabeImage.CMMouseEnter(var msg: TMessage);
begin
   If Assigned(FOnMouseEnter) then
      FOnMouseEnter(Self);
   If fcurson then
      Screen.Cursor := FCursUpHandle
   else
      Screen.Cursor := FCursDisHandle;
end;


procedure TGabeImage.CMMouseLeave(var msg: TMessage);
begin
   if Assigned(FOnMouseLeave) then
      FOnMouseLeave(Self);
   Screen.Cursor := FCursDefHandle;
end;


procedure TGabeImage.SetCursHandle(Index: Integer;const Value: Integer);
begin
   Case Index of
    0:If fcursdnhandle<>Value then FCursDnHandle := Value;
    1:If fcursuphandle<>Value then FCursUpHandle := Value;
    2:If fcursdefhandle<>Value then FCursDefHandle := Value;
    3:If fcursdishandle<>Value then FCursDisHandle := Value;
   end;
   Invalidate;
end;

function TGabeImage.GetCursHandle(Index:Integer):integer;
begin
   Case Index of
    0:Result:=FCursDnHandle;
    1:Result:=FCursUpHandle;
    2:Result:=FCursDefHandle;
    3:Result:=FCursDisHandle;
   end;
end;

procedure TGabeImage.MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
   If enabled then
    begin
     Screen.Cursor := FCursDnHandle;
     Invalidate;
    end;
end;

procedure TGabeImage.MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
   If enabled then
    begin
     Screen.Cursor := FCursUpHandle;
     Invalidate;
    end;
end;

procedure TGabeImage.SetTransparent(const Value: Boolean);
begin
  FTransparent := Value;
  Invalidate;
end;

procedure TGabeImage.SetCursOn(const Value: Boolean);
begin
  IF FCurson<>Value then FCurson := Value;
end;

end.
