unit GabeLabel;

interface

{$A+}{$B-}{$D-}{$L-}{$R-}{$S-}{$T-}

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, DsgnIntf;

type
  TOnMouseEnter = Procedure(Sender: TObject) of Object;
  TOnMouseLeave = Procedure(Sender: TObject) of Object;
  TCaptionStyle = (csNone, csRaised, csHeavyRaised, csRecessed, csHeavyRecessed);
  TGabeLabel = class(TLabel)
  private
    FOnMouseEnter: TOnMouseEnter;
    FOnMouseLeave: TOnMouseLeave;
    FCursDnHandle: Integer;
    FCursUpHandle: Integer;
    FCursDefHandle: Integer;
    FCursDisHandle: Integer;
    FCurson: boolean;
    FCaptionStyle: TCaptionStyle;
    procedure SetCursHandle(Index: Integer;const Value: Integer);
    function  GetCursHandle(Index: Integer): Integer;
    procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MouseUp(Sender: TObject; Button: TMouseButton;  Shift: TShiftState; X, Y: Integer);
    procedure CMMouseEnter(var msg:TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
    procedure SetCurson(const Value: boolean);
    procedure DoDrawText(var Rect: TRect; Flags: Word);
  protected
    procedure Paint; override;
    procedure SetCaptionStyle(Value: TCaptionStyle);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property CursDnHandle: Integer index 0 read GetCursHandle write SetCursHandle;
    property CursUpHandle: Integer index 1 read GetCursHandle write SetCursHandle;
    property CursDefHandle: Integer index 2 read GetCursHandle write SetCursHandle;
    property CursDisHandle: Integer index 3 read GetCursHandle write SetCursHandle;
    property Curson: boolean read FCurson write SetCurson default True;
    property OnMouseEnter: TOnMouseEnter read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TOnMouseLeave read FOnMouseLeave write FOnMouseLeave;
    property CaptionStyle: TCaptionStyle read FCaptionStyle write SetCaptionStyle default csRecessed;
  end;

procedure Register;

implementation

procedure TGabeLabel.SetCursHandle(Index: Integer;const Value: Integer);
begin
   Case Index of
    0:If fcursdnhandle<>Value then FCursDnHandle := Value;
    1:If fcursuphandle<>Value then FCursUpHandle := Value;
    2:If fcursdefhandle<>Value then FCursDefHandle := Value;
    3:If fcursdishandle<>Value then FCursDisHandle := Value;
   end;
   Invalidate;
end;

function TGabeLabel.GetCursHandle(Index:Integer):integer;
begin
   Case Index of
    0:Result:=FCursDnHandle;
    1:Result:=FCursUpHandle;
    2:Result:=FCursDefHandle;
    3:Result:=FCursDisHandle;
   end;
end;

procedure TGabeLabel.MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
   If enabled then
    begin
     Screen.Cursor := FCursDnHandle;
     Invalidate;
    end;
end;

procedure TGabeLabel.MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
   If enabled then
    begin
     Screen.Cursor := FCursUpHandle;
     Invalidate;
    end;
end;

procedure TGabeLabel.CMMouseEnter(var msg: TMessage);
begin
   If Assigned(FOnMouseEnter) then
      FOnMouseEnter(Self);
   If fcurson then
      Screen.Cursor := FCursUpHandle
   else
      Screen.Cursor := FCursDisHandle;
end;


procedure TGabeLabel.CMMouseLeave(var msg: TMessage);
begin
   if Assigned(FOnMouseLeave) then
      FOnMouseLeave(Self);
   Screen.Cursor := FCursDefHandle;
end;

procedure TGabeLabel.SetCurson(const Value: boolean);
begin
  IF FCurson<>Value then FCurson := Value;
end;

constructor TGabeLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 16;
  Width := 62;
  FCurson:=true;
end;

procedure TGabeLabel.SetCaptionStyle(Value: TCaptionStyle);
begin
  if Value <> FCaptionStyle then
   begin
    FCaptionStyle := Value;
    Invalidate;
   end;
end;

procedure TGabeLabel.DoDrawText(var Rect: TRect; Flags: Word);
var
  Text       : array[ 0..255 ] of Char;
  TmpRect    : TRect;
  UpperColor : TColor;
  LowerColor : TColor;
begin
  GetTextBuf(Text, SizeOf(Text));
  if (Flags and DT_CALCRECT <> 0) and ((Text[0] = #0) or ShowAccelChar
      and (Text[0] = '&') and (Text[1] = #0)) then
    StrCopy(Text, ' ');

  if not ShowAccelChar then Flags := Flags or DT_NOPREFIX;

  Canvas.Font := Font;

  UpperColor := clBtnHighlight;
  LowerColor := clBtnShadow;

  if (FCaptionStyle = csRecessed) or (FCaptionStyle = csHeavyRecessed) then
   begin
    UpperColor := clBtnShadow;
    LowerColor := clBtnHighlight;
   end;

  if FCaptionStyle in [csRaised, csHeavyRaised, csRecessed, csHeavyRecessed] then
   begin
    If FCaptionStyle in [csRecessed, csHeavyRecessed, csHeavyRaised] then
     Begin
      TmpRect := Rect;
      OffsetRect( TmpRect, 1, 1 );
      Canvas.Font.Color := LowerColor;
      DrawText(Canvas.Handle, Text, StrLen(Text), TmpRect, Flags);
     End;
    If FCaptionStyle in [csRaised, csHeavyRecessed, csHeavyRaised] then
     Begin
      TmpRect := Rect;
      OffsetRect( TmpRect, -1, -1 );
      Canvas.Font.Color := UpperColor;
      DrawText(Canvas.Handle, Text, StrLen(Text), TmpRect, Flags);
     End;
   end;

  Canvas.Font.Color := Font.Color;
  if not Enabled then Canvas.Font.Color := clGrayText;
  DrawText(Canvas.Handle, Text, StrLen(Text), Rect, Flags);
end;

procedure TGabeLabel.Paint;
const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  Rect: TRect;
begin
  with Canvas do
   begin
    if not Transparent then
     begin
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(ClientRect);
     end;
    Brush.Style := bsClear;
    Rect := ClientRect;
    DoDrawText( Rect, ( DT_EXPANDTABS or DT_WORDBREAK ) or Alignments[ Alignment ] );
  end;
end;

procedure Register;
begin
  RegisterComponents( 'Gabes', [TGabeLabel] );
end;

end.
