unit MiniMax;

{$B-}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, Buttons, Menus, Spin, ComCtrls,
  SchachProcs;

const

  maximal = 20; {hchste Solltiefe}

  BRETT_DIM =       120;  { Dimension des erweiterten Schachbrettes  }
  MAX_TIEFE =       30;  { Maximale Suchtiefe                       }{INFO - war vorher 20}
  ZUGRICHTUNGEN =    16;  { Anzahl der Zugrichtungen fuer alle Fig.  }
  FIGTYPEN =          7;  { Anzahl der Figuretypen - Bezogen auf     }
                            { Zugrichtungen. (wDame == sDame)          }
  ZSTAPEL_DIM =     1000; { Dimension des Zugstapels.                }
  {*
  *** Werte der Reihen und Linien
  *}
  A_LINIE =  1;
  B_LINIE =  2;
  C_LINIE =  3;
  D_LINIE =  4;
  E_LINIE =  5;
  F_LINIE =  6;
  G_LINIE =  7;
  H_LINIE =  8;
  REIHE_1 =  2;
  REIHE_2 =  3;
  REIHE_3 =  4;
  REIHE_4 =  5;
  REIHE_5 =  6;
  REIHE_6 =  7;
  REIHE_7 =  8;
  REIHE_8 =  9;

  {*
  *** Material-Wert der Figuren
  *}
  MAT_B =  100;
  MAT_T =  500;
  MAT_L =  350;
  MAT_S =  325;
  MAT_D =  900;
  MAT_K =   0;     { Nachdem beide Seiten genau einen Koenig haben, }
                  { kann man den Wert auf 0 setzen.                }

 {*
 *** Bewertung fuer Matt
 *}
  MATTWERT = 32000;
  MAX_POS  =  MAT_L;     { Maximum der Positionsbewertung }

  {*
  *** Bonus fuer Hauptvariante und Killer Zuege. Werden fuer Zugsortierung
  *** verwendet.
  *}
  HVARBONUS =      500;
  KILLER1BONUS =   250;
  KILLER2BONUS =   150;

  {*
  *** Gesamtmaterial in der Grundstellung.
  **}
  MAT_GESAMT = ((4*(MAT_T+MAT_L+MAT_S))+(2*MAT_D));
  ENDSPIELMATERIAL =(4*MAT_T+(2*MAT_L));

  {*
  *** Symbolische Konstanten fuer logische Zustaende/Abfragen.
  *}
  WAHR          = 1;
  FALSCH        = 0;
  LEGAL         = 1;
  ILLEGAL       = 0;
  OK            = 1;


type

  zug_t = record
    von : integer;         { Von-Feld           }
    nach : integer;        { Nach-Feld          }
    geschl_fig : integer;  { geschlagene Figur  }
    umwandl_fig : integer; { umgewandelte Figur }
    rochade_nr : integer;  { Art der Rochade    }
    ep_feld : integer;     { Enpassant Feld     }
    wert : integer;        { Bewertung fuer Zugsortierung }
  end;


  {*
  *** Index der Figuren in Offset-Liste und Lang/Kurzschritt Kennung.
  *** Wird vom Zuggenerator verwendet.
  *}
  figoffset_t = record
    anfang : integer;
    ende : integer;
    langschritt : integer;
  end;


  {*
  *** Information ueber Bauern/Figuren Konstellationen.
  *}
  beidefarben_t = record
    weiss : integer;
    schwarz : integer;
  end;

  {*
  *** Information ueber von/nachfeld (Zug ohne zugehoerige Information)
  *** Wird fuer die Speicherung von aussichtsreichen Zuegen (Hauptvariante,
  *** Killerzuege) verwendet.
  *}
  vonnach_t = record
    von : integer;
    nach : integer;
  end;

  {*
  *** Datenstruktur zur Speicherung von 'Killer' Zuegen.
  *}
  killer_t = record
    killer1 : vonnach_t;
    killer2 : vonnach_t;
  end;

  TZugStapel = array[0..ZSTAPEL_DIM] of zug_t;
  PZugStapel = ^TZugStapel;

  TStopEvent = procedure (Sender: TObject; var Zug : PZug; Status : Byte) of object;
  TTraceEvent = procedure (Sender: TObject; S : String) of object;

  TMiniMax = class(TCustomControl)
    constructor Create(AComponent : TComponent);override;
    destructor  Destroy; override;
  private
    FEnglish : Boolean;
    FOnStop : TStopEvent;
    FOnTrace : TTraceEvent;
    FOnSchach : TNotifyEvent;

    slLines : TStringList;
    hilf : string;
    Retttiefe : Integer;
    status : byte;
    MMZug : TZug;
    in_Vorschlag : boolean;
    rpt3S : array[1..2] of TSchachbrett; {soll Zugwiederholung in besserer Stellung verhindern (Proz. akzeptabel)}
    rpt3Z : array[1..2] of Zug_t;

    MMBrett : TSchachBrett;
    Ep_Feld : array[0..MAX_TIEFE-1] of integer;

    ZugStapel : PZugStapel;

    Bewegt : array[0..H8] of integer;  { Zaehlt mit, wie oft sich eine Figur von Feld}
                                           { ** bewegt hat. Wird fuer Ueberpruefung der Rochade}
                                           { ** verwendet (auch fuer Bewertung nuetzlich).}

    Rochade : array[0..3] of integer;    { Hat WEISS/SCHWARZ schon rochiert? }
    Index : integer;         { Index in Zugstapel. }

    {*
    *** Notiert Position auf Zugstapel. Zuege der Tiefe 'n' werden im Bereich
    *** [ StapelGrenze[n], StapelGrenze[n+1] ) auf Zugstapel abgelegt.
    *}
    StapelGrenze : array[0..MAX_TIEFE-1] of integer;
    HVar : array[0..MAX_TIEFE-1,0..MAX_TIEFE-1] of vonnach_t;  { Hauptvariante Tabelle.   }
    KillerTab : array[0..MAX_TIEFE-1] of killer_t;        { Killer-Zuege Tabelle.    }

    {*
    *** Tabellen fuer Bewertungsfunktion
    **}
    BKontrolle : array[0..BRETT_DIM-1] of beidefarben_t; { Felder die von Bauern kontrolliert }
                                                         { werden. }
    Bauern : array[0..H_LINIE+2-1] of beidefarben_t;  { Anzahl der Bauern auf einer Linie }
    Tuerme : array[0..H_LINIE+2-1] of beidefarben_t;  { Anzahl der Tuerme auf einer Linie }
    Mobilitaet : array[0..MAX_TIEFE-1] of integer;   { Mobilitaet von Laeufer und Turm   }
    NachFeld : array[0..MAX_TIEFE-1] of integer;     { NachFeld der Zuege. Wird zur Zug- }
                                                     { sortierung und zur Erweiterung der }
                                                     { Suche verwendet.                   }
    wKoenig,sKoenig : Integer;            { Position des weissen und schwarzen Koenigs }
    MatBilanz : array[0..MAX_TIEFE-1] of Integer; { Material Bilanz zwischen Weiss und Schwarz }
    MatSumme : array[0..MAX_TIEFE-1] of Integer;  { Gesamtmaterial am Brett.                   }
    Farbe : Integer;                { Wer ist am Zug. }
    SollTiefe : Integer;            { Suche wird bis Solltiefe durchgefuehrt.    }
    MaxErweiterung : Integer;       { Erweiterungen im Suchbaum auf Grund von    }
                                    { Schach oder Schlagen werden nur bis Max-   }
                                    { Erweiterungen durchgefuehrt. (Ansonsten    }
                                    { kann Suche explodieren).                   }
    Tiefe : Integer;                { Suchtiefe. Anzahl der Halbzuege von der    }
                                    { Ausgangsstellung.                          }
    KnotenAnzahl : longInt;         { Anzahl der untersuchten Stellungen/Knoten. }
    LetzterZug : Integer;           { Zuletzt ausgefuehrter Zug }
    InSchach : Integer;             { Seite am Zug steht im Schach }
    ZugNr : Integer;                { Bisher im Spiel ausgefuehrte Halbzuege     }

    FNoCheckMessages : Boolean;
    FVorschlagtiefe : byte;
    FStop : Boolean;
    FBreak : Boolean;

    FFRC : PFischerInfo;

    PROCEDURE initialisiere;
    PROCEDURE stellungs_eingabe ;
    PROCEDURE computer_zug ;
    PROCEDURE generiere_zuege(allezuege : integer);
    PROCEDURE init_spielbaum;
    FUNCTION  zug_eingabe(von,nach, UFigur : integer; FiRochade, NoMoveGen : Boolean) : integer;
    FUNCTION  feld_notation(feldnummer : integer; feldnotation : string) : string;
    FUNCTION  zeige_zug(aktzug : integer) : String;
    PROCEDURE zug_ausfuehren(aktzug : integer);
    PROCEDURE zug_zuruecknehmen(aktzug : integer);
    FUNCTION  bewerte_stellung(alpha, beta, seite : integer) : integer;
    PROCEDURE init_bewertung ;
    PROCEDURE kopiere_hvar(aktzug : integer);
    FUNCTION  sgn(wert : integer) : Integer;
    FUNCTION  naechstbesterzug : integer;
    FUNCTION  alpha_beta(alpha, beta, distanz : integer)  : integer;
    PROCEDURE zeige_hvar;
    FUNCTION  sbauerbewertung(feld, reihe,  linie, entwickelt : integer) : integer;
    FUNCTION  wbauerbewertung(feld, reihe, linie, entwickelt : integer) : integer;
    FUNCTION  attackiert_feld(feld,  seite : integer ) : integer;
    PROCEDURE notiere_zug( von, nach : integer);
    PROCEDURE notiere_schlagzug( von, nach : integer);
    PROCEDURE notiere_epzug( von,nach, epfeld : integer);
    PROCEDURE notiere_umwandlung(von, nach : integer);
    procedure StellungsDatenEintragen;
    procedure akzeptabel;

    procedure PStop;
  protected
    procedure Paint; override;
    procedure Loaded; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  public
    property  NoCheckMessages : Boolean read FNoCheckMessages write FNoCheckMessages;
    property  Vorschlagtiefe : Byte read FVorschlagtiefe write FVorschlagtiefe;
    property  Stop : Boolean read FStop write FStop;
    property  FRC : PFischerInfo read FFRC write FFRC;
    procedure BreakLoop;
    procedure Start(stellung : TSchachbrett;Zug : TZug; t:byte);
    function  IstZugLegal(stellung : TSchachbrett; Zug : TZug) : Boolean;
    function  Vorschlag(stellung : TSchachbrett;Zug : TZug; feld : byte) : byte;
    function  aktuellerWert : real;    {akt. Stellungsbew. kann abgefragt werde, damit Berechnung geg.
    aus Zeitgrnden unterbrochen werden kann (oder verlngert)}
  published
    property English : Boolean read FEnglish write FEnglish;
    property OnStop : TStopevent read FOnStop write FOnstop;
    property OnTrace : TTraceevent read FOnTrace write FOnTrace;
    property OnSchach : TNotifyEvent read FOnSchach write FOnSchach;
  end;


implementation

const
Offset : array[0..ZUGRICHTUNGEN-1] of integer = (
   -9,-11,9,11,                  { Laeufer Richtungen }
   -1,10,1,-10,                  { Turm               }
   19,21,12,-8,-19,-21,-12,8);   { Springer           }

FigOffset : array[0..FIGTYPEN-1] of figoffset_t= (
{*
*** Erster Wert ist Offset-Anfang, zweiter Offset-Ende, dritter Langzug
*}
   (anfang:0;ende:0;langschritt:FALSCH),  { Leeres Feld                       }
   (anfang:-1;ende:-1;langschritt:FALSCH),{ Bauernzuege werden extra erzeugt. }
   (anfang:4;ende:7;langschritt:WAHR),    { Turm                              }
   (anfang:0;ende:3;langschritt:WAHR),    { Laeufer                           }
   (anfang:8;ende:15;langschritt:FALSCH), { Springer                          }
   (anfang:0;ende:7;langschritt:WAHR),    { Dame                              }
   (anfang:0;ende:7;langschritt:FALSCH)); { Koenig                            }

{*
*** Zentralisations-Tabelle. Wir brauchen nur die Felder 0..H8, da
*** Figur nicht auf einem Feld groeszer H8 stehen kann.
*** Unterer Rand wird beibehalten, da man sonst von Brettkoordinate in
*** Zentralitaets-Koordinate umrechnen muss.
*** H1 ist vom Zentrum weiter weg als G1. Trotzdem bekommt H1 besseren
*** Zentrumswert als G1. Diese Tabelle wird u.a. fuer Koenigsbewertung
*** verwendet. Die Werte von G1,H1 bewirken, dass Koenig nach Rochade
*** auf G1 stehen bleibt und nicht den unnoetigen Zug G1-H1 ausfuehrt.
*** (Der Springer steht weder auf G1,noch H1 besonders gut).
*}
const ZentrumTab : array[0..H8] of integer = (
{*    A   B   C   D   E   F   G   H     *}
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  4,  0,  8, 12, 12,  8,  0,  4,  0,   { A1 ... H1 }
   0,  4,  8, 12, 16, 16, 12,  8,  4,  0,   { A2 ... H2 }
   0,  8, 12, 16, 20, 20, 16, 12,  8,  0,   { A3 ... H3 }
   0, 12, 16, 20, 24, 24, 20, 16, 12,  0,   { A4 ... H4 }
   0, 12, 16, 20, 24, 24, 20, 16, 12,  0,   { A5 ... H5 }
   0,  8, 12, 16, 20, 20, 16, 12,  8,  0,   { A6 ... H6 }
   0,  4,  8, 12, 16, 16, 12,  8,  4,  0,   { A7 ... H7 }
   0,  4,  0,  8, 12, 12,  8,  0,  4);      { A8 ... H8 }


{*
*** Bewertung der Felder fuer Bauern.
*** Wird in Stellungsbewertung verwendet.
*}
const wBFeldWert : array[0..H7] of integer= (   { weisse Bauern }
{*    A   B   C   D   E   F   G   H     *}
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,  { A1..H1 (nicht moeglich) }
   { Zentrumsbauern sind auf 2-ter Reihe schlecht (gehoeren nach vorne) }
   { f-h Bauern sollen wegen Koenigsschutz hinten bleiben }
   0,  4,  4,  0,  0,  0,  6,  6,  6, 0,
   0,  6,  6,  8,  8,  8,  4,  6,  6, 0,   { A3 ... H3 }
   0,  8,  8, 16, 22, 22,  4,  4,  4, 0,   { A4 ... H4 }
   0, 10, 10, 20, 26, 26, 10, 10, 10, 0,   { A5 ... H5 }
   0, 12, 12, 22, 28, 28, 14, 14, 14, 0,   { A6 ... H6 }
   0, 18, 18, 28, 32, 32, 20, 20, 20);     { A7 ... H7 }
   { Auf 8-ter Reihe kann kein Bauer stehen }

const sBFeldWert : array[0..H7] of integer= (    { schwarze Bauern }
{*    A   B   C   D   E   F   G   H     *}
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,  { A1..H1 (nicht moeglich) }
   0, 18, 18, 28, 32, 32, 20, 20, 20, 0,  { A2 ... H2 }
   0, 12, 12, 22, 28, 28, 14, 14, 14, 0,  { A3 ... H3 }
   0, 10, 10, 20, 26, 26, 10, 10, 10, 0,  { A4 ... H4 }
   0,  8,  8, 16, 22, 22,  4,  4,  4, 0,  { A5 ... H5 }
   0,  6,  6,  8,  8,  8,  4,  6,  6, 0,  { A6 ... H6 }
   0,  4,  4,  0,  0,  0,  6,  6,  6);    { A7 ... H7 }   { Auf 8-ter Reihe kann kein Bauer stehen }

{*
*** Material-Wert der Figuren
**}
const FigMaterial : array[0..FIGTYPEN-1] of integer= (0,MAT_B,MAT_T,MAT_L,MAT_S,MAT_D,MAT_K);

const FigSymbol : array[0..FIGTYPEN-1] of Char = ( '.','B','T','L','S','D','K' );

const
vokabeln : array[0..3] of string[10] =   {fr Deutsch u. Englisch in MiniMax}
('Tiefe','Depth','Wert','Value');

var
  aBitmap: TBitmap;
  Loaded: Boolean;

procedure LoadBitmap;
begin
  if Loaded then exit;
  Loaded := True;
  if aBitmap = nil then aBitmap := TBitmap.Create;
  try
    aBitmap.LoadFromResourceName(HInstance, 'TMINIMAX');
  except
    on E:Exception do ShowMessage(E.Message);
  end;
end;

procedure TMiniMax.Paint;
begin
  if csDesigning in ComponentState then
    Canvas.StretchDraw(Canvas.ClipRect, aBitmap);
  inherited Paint;
end;


procedure TMiniMax.Loaded;
begin
  if csDesigning in ComponentState then LoadBitmap
  else SetBounds(0,0,0,0);
  inherited Loaded;
end;

procedure TMiniMax.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if csDesigning in ComponentState then
  begin
    AWidth := 25;
    AHeight := 25;
  end;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

{--------------------------------------------------------------------}
{  Initialisiert das Brett und den Spielstatus.                      }
{--------------------------------------------------------------------}

PROCEDURE
TMiniMax.initialisiere;
VAR i : integer;
BEGIN
   {*
   *** Brett Initialisierung
   *}
   ZugNr:=0;
   for i:=0 to BRETT_DIM-1 do BEGIN
      MMBrett[i] := AusgangsStellung[i];
   END;
   { Position der Koenige in der Grundstellung }
   wKoenig:=E1;
   sKoenig:=E8;
   { Noch keine Rochade ausgefuehrt. }
   for i:=0 to 2 do BEGIN
      Rochade[i]:=FALSCH;
   END;
   for i:=A1 to H8 do BEGIN
      Bewegt[i]:=0;    { Zu Beginn hat sich noch keine Figur bewegt }
   END;
   Ep_Feld[0] := ILLEGAL;     { Enpassant Status }
   MatSumme[0] := MAT_GESAMT; { Material (der Figuren) in Grundstellung   }
   MatBilanz[0]:=0;           { Ausgeglichene Materialbilanz.             }
   StapelGrenze[0]:=0;     { Grenze des Zugstapels       }
   SollTiefe:=4;           { Voreinstellung Rechentiefe  }
   Tiefe    :=0;           { Aktuelle Tiefe im Spielbaum }
   RettTiefe:=0;
   Farbe    :=WEISS;       { Weiss in Grundstellung am Zug }
   Stop:=true;
END;

{--------------------------------------------------------------------}
{- Signum-Funkion: 'Wert' des Vorzeichens.                          -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.sgn(wert : integer) : Integer;
BEGIN
   if(wert=0) then BEGIN
      sgn:= 0;
   END
   else if(wert>0) then BEGIN
      sgn:= 1;
   END
   else sgn:= -1;
END;
{--------------------------------------------------------------------}
{- Verwandelt interne Feld Nummer in Feld Notation.                 -}
{- Falls die Nummer nicht am Brett liegt, wird '--' zurueck gegeben -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.feld_notation(feldnummer : integer;feldnotation : string) : string;
BEGIN
   { Ueberpruefe Korrektheit }
   if((feldnummer<A1)OR(feldnummer>H8)OR(MMBrett[feldnummer]=RAND)) then BEGIN
      feldnotation:='--';
   END
   else BEGIN
      feldnotation:=chr(65- 1 + feldnummer mod 10); { '%' ist Modulo Operator }
      feldnotation:=feldnotation+chr(trunc(49- 2 + (feldnummer/10)));
   END;
    feld_notation:=feldnotation;
END;

{--------------------------------------------------------------------}
{- Eingabe einer beliebigen Stellung.                               -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.stellungs_eingabe;
var
    i : integer;
    wkoenige,skoenige : integer;   { Anzahl der weissen und schwarzen Koenige }
label   continue;
BEGIN
   Tiefe:=0;         { Stellung wird Wurzel des Suchbaumes }
   {*
   *** Berechne Material Bilanz und ueberpruefe ob jede Seite genau
   *** einen Koenig hat. 9-Bauern und 17 Figuren sind aber 'legal'.
   *}
   MatBilanz[0]:=0;
   MatSumme[0]:=0;
   wkoenige:=0;
   skoenige:=0;
   for i:=A1 to H8 do BEGIN  { Lese jedes Feld }
      if((MMBrett[i]=LEER)OR(MMBrett[i]=RAND)) then BEGIN
         goto continue; { Leer- oder Randfeld gefunden, gehe zum naechsten Feld }
      END;
      { Neue MaterialBilanz }
      { Weisser Stein geht in Bilanz positiv ein, schwarzer Negativ }
      MatBilanz[0]:=MatBilanz[0] + sgn(MMBrett[i])*FigMaterial[abs(MMBrett[i])];
      if(abs(MMBrett[i])<>W_B) then BEGIN
         MatSumme[0]:=MatSumme[0] + FigMaterial[abs(MMBrett[i])];
      END;
      if MMBrett[i]=W_K then wkoenig:=i;
      if MMBrett[i]=S_K then skoenig:=i;
      continue:
   END;

   if Farbe=SCHWARZ then              { Material Bilanz wurde bisher aus der }
      MatBilanz[0] := -MatBilanz[0];  { der Sicht von WEISS berechnet.       }

   for i:=A1 to H8 do BEGIN          { Eingebene Stellungs meistens aus     }
      Bewegt[i]:=1;                  { Mittel- oder Endspiel. Setze Bewegt  }
                                    { auf 1, entsprechende Teile der Be-   }
				    { wertungsfunktion aufgerufen wird.    }
   END;
   if assigned(FRC) then with FRC^ do begin
     Bewegt[FischerWhiteKing] := 0;
     Bewegt[FischerBlackKing] := 0;
   end
   else begin
     Bewegt[E1]:=0;                    { Koenigfelder stellen Rochade    }
     Bewegt[E8]:=0;                  { Status dar. Muessen daher auf 0      }
   end;
                                   { gesetzt werden.                      }
   { Ruecksetzen der Zugnummer. }
   ZugNr:=0;
END;

{--------------------------------------------------------------------}
{- Initialisiere den Spielbaum.                                     -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.init_spielbaum;
BEGIN
   if(Tiefe=0) then BEGIN { Noch nichts gerechnet, Spielbaum bereits initialisiert }
      exit;
   END;
   Ep_Feld[0] := Ep_Feld[1];
   MatBilanz[0]:=MatBilanz[1];
   MatSumme[0]:=MatSumme[1];
   Tiefe := 0;
END;
{--------------------------------------------------------------------}
{- Aus den moeglichen Zuegen einer bestimmten Tiefe wird der beste  -}
{- noch nicht gespielte Zug ausgewaehlt. Gibt den Index des Zuges   -}
{- am Zugstapel zurueck. Bereits alle Zuege gespielt wurden, wird   -}
{- ein unmoeglicher Index (-1) zurueck gegeben.                     -}
{- Der Wert eines Zuges wird vom Zuggenerator bestimmt.             -}
{- Diese Funktion erledigt die Zugsortierung in der Suche.          -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.naechstbesterzug : integer;
Var i,besterwert,besterzug : integer;
BEGIN
   besterzug:=-1;
   besterwert:=-MATTWERT;
   i:=StapelGrenze[Tiefe];
   while i< StapelGrenze[Tiefe+1] do BEGIN
      if(ZugStapel^[i].wert>besterwert) then BEGIN { Neuer bester Zug gefunden }
         besterzug:=i;
         besterwert:=ZugStapel^[i].wert;
      END;
      inc(i);
   END;
   { Markiere ausgewaehlten Zug, damit er beim naechsten Aufruf nicht }
   { mehr ausgewaehlt wird.                                           }
   if(besterzug>=0) then BEGIN
      ZugStapel^[besterzug].wert:=-MATTWERT;
   END;
   naechstbesterzug:= besterzug;
END;
{---------------------------------------------------------------------}
{- Alpha-Beta Baumsuche.                                             -}
{- Gibt Bewertung aus der Sicht der Farbe die am Zug ist zurueck.    -}
{- 'alpha' ist untere Schranke, 'beta' obere Schranke und 'distanz'  -}
{- die Anzahl der Halbzuege bis zum Horizont.                        -}
{- Ist 'distanz' positiv, wird normale alpha-beta Suche ausgefuehrt, -}
{- im Falle kleiner gleich 0 die Ruhesuche.                          -}
{- Es wird der NegaMax-Wert aus der Sicht der Seite am Zug zurueck   -}
{- gegeben.                                                          -}
{- Diese Prozedure wird rekursiv aufgerufen.                         -}
{---------------------------------------------------------------------}
FUNCTION
TMiniMax.alpha_beta(alpha, beta, distanz : integer)  : integer;
var
   i,wert,besterwert,schach : integer;
label fertig;
BEGIN
   KnotenAnzahl:=KnotenAnzahl+1;  { Zusaetzliche Stellung untersucht.   }
   HVar[Tiefe][Tiefe].von:=0;     { Loesche bisherige Hauptvariante     }
   { Stellung wird immer, d.h. auch im Inneren des Baumes bewertet.    }
   { Dies ist wegen Matt- und Patt-Erkennung notwendig. Weiters        }
   { wird die Bewertung zur Steuerung der Suche/Erweiterung verwendet. }
   { Die Knotenanzahl im Inneren des Baumes ist viel kleiner als jene  }
   { am Horizont. D.h. das Programm wird dadurch nicht wesentlich      }
   { langsamer.                                                        }
   {                                                                   }
   { Bewertung aus der Sicht der Seite die am Zug ist.                 }
   {                                                                   }
   wert:=bewerte_stellung(alpha,beta,Farbe);  
   {                                                                   }
   { Im Falle eines Schachs wird Suche erweitert. Maximal um insgesamt }
   { 4-Halbzuegen. Ansonsten kann es passieren, dass der Suchbaum      }
   { bei gegenseitigen Schach und Schlagsequenzen extrem gross wird.   }
   { In der Regel sind diese Sequenzen vollkommen sinnlos.             }
   {  !!! (auch For Schleifen prfen - durch while ersetzen            }
   schach:=InSchach;
   if((schach=WAHR)AND(Tiefe+distanz<MaxErweiterung+1)) then BEGIN
      distanz:=distanz+1;  { Seite am Zug steht im Schach. Erweitere Suche }
   END
   {                                                                   }
   { Bei Schlagen und Rueckschlagen auf dasselbe Feld wird Suche       }
   { erweitert, falls die Materialbilanz ungefaehr im Gleichgewicht    }
   { bleibt und ausserdem bisher nocht nicht zuviele Erweiterungen     }
   { gemacht wurden.                                                   }
   {                                                                   }
   else if((Tiefe>=2)AND(Tiefe+distanz<MaxErweiterung)AND
      (NachFeld[Tiefe]=NachFeld[Tiefe-1])AND
      (wert>=alpha-150)AND(wert<=beta+150)) then BEGIN { Materialbilanz ungef. gleich }
      distanz:=distanz+1;
   END;
   if (not NoCheckMessages) and (not in_Vorschlag) and (not stop) then Application.ProcessMessages;
   if Stop then begin
     alpha_beta:= wert;
     exit;
   end;
   { 
   ** Falls bereits mehr als 5-Zuege in der Ruhesuche ausgefuehrt wurden,
   ** oder der Gegner Matt ist oder auf Grund der Datenstrukturen maximale
   ** Suchtiefe erreicht ist, beende Suche. 
   }
   if((distanz<-5)OR(wert=MATTWERT-Tiefe)OR(Tiefe>=MAX_TIEFE)) then BEGIN
      alpha_beta:= wert;
      exit;
   END;
   {* 
   *** Falls in der Ruhesuche die Seite am Zug gut steht, wird Suche
   *** abgebrochen, da per Definition der Wert in der Ruhesuche nur
   *** besser werden kann.
   *** Achtung: Es wird auch bereits bei distanz 1, d.h. einen Halbzug
   ***          vor dem Horizont abgebrochen, falls Seite am Zug nicht
   ***          im Schach steht. Dies ist eine selektive Abweichung vom
   ***          brute-force alpha-beta Schema.
   *}
   if((wert>=beta)AND(distanz+schach<=1)) then BEGIN
      alpha_beta:=wert;
      exit;
   END;
   {*
   *** Erzeuge Zuege. Falls distanz <:= 0 ist (Ruhesuche) werden nur
   *** Schlagzuege und Umwandlungen erzeugt.
   *}
   generiere_zuege(distanz); { Abfrage, ob ueberhaupt Zuege vorhanden }
   if(distanz>0) then BEGIN           { sind, wird indirekt von der Festlegung }
      besterwert := -MATTWERT; { von besterwert mit erledigt.          }
   END
   else BEGIN   { In Ruhesuche ist die aktuelle Stellungsbewertung untere }
      besterwert := wert;  { Schranke fuer den Suchwert.               }
   END;
   { Untersuche alle Zuege in sortierter Reihenfolge.}
   i:=naechstbesterzug;
   while (i>=0) do BEGIN
      zug_ausfuehren(i);
      { NegaMax Prinzip. Vorzeichen wird umgedreht, Rolle von Alpha, Beta
      ** vertauscht.
      *}
      wert:=-alpha_beta(-beta,-alpha,distanz-1);
      zug_zuruecknehmen(i);
      if(wert>besterwert) then BEGIN  { neuer bester Wert gefunden }
         besterwert:=wert;
         if(wert>=beta) then BEGIN    { Cutoff gefunden }
            if(distanz>0) then BEGIN    { Im Inneren des Baumes wird noch neue }
              kopiere_hvar(i); { Hauptvariante notiert.               }
            END;
            goto fertig;   
         END;
         if(wert>alpha) then BEGIN  { Wert ist verbesserte untere Schranke }
            if(distanz>0) then BEGIN     { Im Inneren des Baumes wird noch neue }
               kopiere_hvar(i); { Hauptvariante notiert.               }
            END;
            alpha:=wert;       { Verbesserter alpha Wert }
         END;
      END;
      i:=naechstbesterzug;
   END; 
fertig:
   { Guter Zug, der Cutoff erzeugt, wird in Killertabelle eingetragen. }
   { Bewahre den bisher besten Killer als zweitbesten Killer auf.      }
   if((wert>=beta)AND(i>=0)) then BEGIN
      KillerTab[Tiefe].killer2:=KillerTab[Tiefe].killer1; 
      KillerTab[Tiefe].killer1.von:=ZugStapel^[i].von;
      KillerTab[Tiefe].killer1.nach:=ZugStapel^[i].nach;
   END;
   { Falls Seite keinen legalen Zug mehr hat }
   if(besterwert= -(MATTWERT-(Tiefe+1))) then BEGIN
      if(schach=FALSCH) then BEGIN  { Aber nicht im Schach steht. }
         alpha_beta:= 0;           { Ist es Patt }
         exit;
      END;
   END;
   alpha_beta:= besterwert;
END;

{--------------------------------------------------------------------}
{- Gibt den aktuellen Zug in Schachnotation aus.                    -}
{- Rochade ist 'E1-G1' und nicht O-O                                -}
{- aktzug ist der Index des Zuges im ZugStapel.                     -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.zeige_zug(aktzug : integer) : String;
var
   feldtext : string[5];
   von,nach : integer;
BEGIN
   von:=ZugStapel^[aktzug].von;
   nach:=ZugStapel^[aktzug].nach;
   { Art der Figur }
   hilf:=(FigSymbol[abs(MMBrett[von])]);
   { Feld }
   hilf:=hilf+(feld_notation(von,feldtext));
   if(ZugStapel^[aktzug].geschl_fig=LEER) then BEGIN
      hilf:=hilf+('-');           { Normaler Zug }
   END
   else BEGIN
      hilf:=hilf+('x');           { Schlagzug   }
   END;
   hilf:=hilf+(feld_notation(nach,feldtext));
   { Falls Umwandlung, fuege Umwandlungs Figur hinzu }
   if(ZugStapel^[aktzug].umwandl_fig<>LEER) then BEGIN
      hilf:=hilf+
      (FigSymbol[ZugStapel^[aktzug].umwandl_fig]);
   END;
   zeige_zug:=hilf+'  ';
END;
{--------------------------------------------------------------------}
{- Traegt den aktuellen Zug in die Hauptvariante ein und kopiert     }
{- die Fortsetzung die auf der naechsten Tiefe gefunden wurde.       }
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.kopiere_hvar(aktzug : integer);
var
   i : integer;
BEGIN
   { Neue Hauptvariante ist Fortsetzung dieser Variante }
   HVar[Tiefe][Tiefe].von:=ZugStapel^[aktzug].von;
   HVar[Tiefe][Tiefe].nach:=ZugStapel^[aktzug].nach;
   i:=1;
   repeat
      if(HVar[Tiefe+1][Tiefe+i].von=0) then BEGIN
         HVar[Tiefe][Tiefe+i]:=HVar[Tiefe+1][Tiefe+i];
         exit;
      END;
      HVar[Tiefe][Tiefe+i]:=HVar[Tiefe+1][Tiefe+i];
      inc(i);
   until 1=2;
END;
{--------------------------------------------------------------------}
{- Anzeige der aktuellen Hauptvariante. Es wird nur von-nach Feld   -}
{- angezeigt.                                                       -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.zeige_hvar;
var
   i : integer;
   vontext, nachtext : string[5];
label weiter;
BEGIN
 if not in_Vorschlag then begin
   sllines[1] := '';
   i:=0;
   repeat
      if(HVar[0][i].von=0) then BEGIN
         goto weiter;
      END;
      sllines[1]:=sllines[1]+
      feld_notation(HVar[0][i].von,vontext)+'-'+
                      feld_notation(HVar[0][i].nach,nachtext)+' ';
      inc(i);
   until 1=2;
   weiter:
 END;
END;
{--------------------------------------------------------------------}
{- Fuehrt einen Zug am Brett aus und aktualisiert den Status und    -}
{- die Suchtiefe.                                                   -}
{- aktzug ist der Index des Zuges im ZugStapel.                     -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.zug_ausfuehren(aktzug : integer);
var
   von,nach,epfeld,matveraenderung : integer;
BEGIN
   ZugNr:=ZugNr+1;   { Zugnummer wird um einen Halbzug erhoeht. }
   von:=ZugStapel^[aktzug].von;
   nach:=ZugStapel^[aktzug].nach;
   epfeld:=ZugStapel^[aktzug].ep_feld;
   LetzterZug:=aktzug;
   Tiefe:=Tiefe+1;                    { Eine Stufe tiefer im Baum }
   NachFeld[Tiefe]:=nach;             { Wird fuer Zugsortierung und Er- }
                                     { weiterung der Suche verwendet.  }
   Ep_Feld[Tiefe]:=ILLEGAL;
   { Material Bilanz wird immer aus der Sicht der Seite am Zug betrachtet. }
   { Daher muss das Vorzeichen umgedreht werden.                           }
   MatBilanz[Tiefe]:= -MatBilanz[Tiefe-1];
   MatSumme[Tiefe]:=MatSumme[Tiefe-1];
   { Figur bewegt sich von von-feld nach nach-feld }
   Bewegt[von]:=Bewegt[von]+1;
   Bewegt[nach]:=Bewegt[nach]+1;
   if(epfeld<>ILLEGAL) then BEGIN
      if(MMBrett[epfeld]=LEER) then BEGIN      { Bauernzug von 2-ter auf 4-te Reihe }
         Ep_Feld[Tiefe]:=epfeld;
      END
      else BEGIN  { Gegnerischer Bauer wird durch enpassant geschlagen }
         MMBrett[epfeld]:=LEER;
         MatBilanz[Tiefe]:= MatBilanz[Tiefe] - MAT_B;
      END;
   END
   else BEGIN  { Falls Figur geschlagen wird, veraendere Material Bilanz   }
      if(ZugStapel^[aktzug].geschl_fig<>LEER) then BEGIN { Figur wurde geschlagen }
         matveraenderung:= FigMaterial[ZugStapel^[aktzug].geschl_fig];
         MatBilanz[Tiefe]:= MatBilanz[Tiefe] - matveraenderung;
         { In Materialsumme geht nur das Material von Figuren ein }
         if(matveraenderung<>MAT_B) then BEGIN
            MatSumme[Tiefe] := MatSumme[Tiefe] - matveraenderung;
         END;
      END;
   END;
   MMBrett[nach]:=MMBrett[von];     { Zug am Brett setzen }
   MMBrett[von]:=LEER;
   if(ZugStapel^[aktzug].umwandl_fig<>LEER) then BEGIN { Bauern Umwandlung }
      MMBrett[nach]:=Farbe*ZugStapel^[aktzug].umwandl_fig;
      matveraenderung:= (FigMaterial[ZugStapel^[aktzug].umwandl_fig]-MAT_B);
      MatBilanz[Tiefe]:=MatBilanz[Tiefe] - matveraenderung;
      { Bauern werden in MatSumme nicht gezaehlt. }
      MatSumme[Tiefe] :=MatSumme[Tiefe] + matveraenderung + MAT_B;
   END
   else BEGIN
      if(ZugStapel^[aktzug].rochade_nr=KURZEROCHADE) then BEGIN
         MMBrett[nach+1]:=LEER; { nach ist G1 oder G8 (je nach Farbe) }
         MMBrett[nach-1]:=Farbe*W_T; { Auf F1/F8 weissen/schwarzen Turm setzen }
         Rochade[Farbe+1]:=WAHR;
      END
      else if(ZugStapel^[aktzug].rochade_nr=LANGEROCHADE) then BEGIN
         MMBrett[nach-2]:=LEER;
         MMBrett[nach+1]:=Farbe*W_T;
         Rochade[Farbe+1]:=WAHR;
      END
      else if(ZugStapel^[aktzug].rochade_nr=KURZEFISCHERROCHADE) then BEGIN
         if not assigned(FRC) then exit;
         if Farbe = WEISS then MMBrett[FRC^.FischerRightWhiteRook]:=LEER
         else MMBrett[FRC^.FischerRightBlackRook]:=LEER;
         MMBrett[nach-1]:=Farbe*W_T; { Auf F1/F8 weissen/schwarzen Turm setzen }
         MMBrett[nach]:=Farbe*W_K;
         Rochade[Farbe+1]:=WAHR;
      END
      else if(ZugStapel^[aktzug].rochade_nr=LANGEFISCHERROCHADE) then BEGIN
         if not assigned(FRC) then exit;
         if Farbe = WEISS then MMBrett[FRC^.FischerLeftWhiteRook]:=LEER
         else MMBrett[FRC^.FischerLeftBlackRook]:=LEER;
         MMBrett[nach+1]:=Farbe*W_T;
         MMBrett[nach]:=Farbe*W_K;
         Rochade[Farbe+1]:=WAHR;
      END;
   END;
   { Falls Koenig gezogen hat, aktualisiere die Position des Koenigs }
   if(MMBrett[nach]=W_K) then BEGIN
      wKoenig:=nach;
   END
   else if(MMBrett[nach]=S_K) then BEGIN
      sKoenig:=nach;
   END;
   { Aendere die Farbe (Seite die am Zug ist) }
   Farbe := -Farbe;
END;
{--------------------------------------------------------------------}
{- Nimmt einen Zug am Brett zurueck.                                -}
{- aktzug ist der Index des Zuges im ZugStapel.                     -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.zug_zuruecknehmen(aktzug : integer);
var
   von,nach,epfeld : integer;
BEGIN
   ZugNr:=ZugNr-1;
   von:=ZugStapel^[aktzug].von;
   nach:=ZugStapel^[aktzug].nach;
   epfeld:=ZugStapel^[aktzug].ep_feld;

   Farbe := -Farbe;
   Tiefe := Tiefe-1;
   MMBrett[von]:=MMBrett[nach];
   MMBrett[nach]:=LEER;
   if((epfeld<>ILLEGAL)AND(ZugStapel^[aktzug].geschl_fig=W_B)) then BEGIN
      MMBrett[epfeld]:=-Farbe; { W_B=WEISS, S_B=SCHWARZ }
   END
   { Geschlagene Figur wieder zurueck setzen }
   else if(ZugStapel^[aktzug].geschl_fig<>LEER) then BEGIN
      MMBrett[nach]:= (-Farbe)*ZugStapel^[aktzug].geschl_fig;
   END;
   { Bewegungszaehler zurueck setzen }
   Bewegt[von]:=Bewegt[von]-1;
   Bewegt[nach]:=Bewegt[nach]-1;
   { Bei Rochade Turm zurueck }
   if(ZugStapel^[aktzug].rochade_nr=KURZEROCHADE) then BEGIN
      MMBrett[nach+1]:=Farbe*W_T;
      MMBrett[nach-1]:=LEER;
      Rochade[Farbe+1]:=FALSCH;
   END
   else if(ZugStapel^[aktzug].rochade_nr=LANGEROCHADE) then BEGIN
      MMBrett[nach-2]:=Farbe*W_T;
      MMBrett[nach+1]:=LEER;
      Rochade[Farbe+1]:=0;
   END
   else if(ZugStapel^[aktzug].rochade_nr=KURZEFISCHERROCHADE) then BEGIN
      if not assigned(FRC) then exit;
      MMBrett[nach-1]:=LEER;
      if Farbe = WEISS then MMBrett[FRC^.FischerRightWhiteRook]:=W_T
      else MMBrett[FRC^.FischerRightBlackRook]:=S_T;
      MMBrett[von]:=Farbe * W_K;
      Rochade[Farbe+1]:=FALSCH;
   END
   else if(ZugStapel^[aktzug].rochade_nr=LANGEFISCHERROCHADE) then BEGIN
      if not assigned(FRC) then exit;
      MMBrett[nach+1]:=LEER;
      if Farbe = WEISS then MMBrett[FRC^.FischerLeftWhiteRook]:=W_T
      else MMBrett[FRC^.FischerLeftBlackRook]:=S_T;
      MMBrett[von]:=Farbe * W_K;
      Rochade[Farbe+1]:=0;
   END;
   if(ZugStapel^[aktzug].umwandl_fig<>LEER) then BEGIN
      MMBrett[von]:=Farbe;  { Bauernumwandlung zurueck nehmen }
   END;
   { Falls Koenig gezogen hat, aktualisiere die Position des Koenigs }
   if(MMBrett[von]=W_K) then BEGIN
      wKoenig:=von;
   END
   else if(MMBrett[von]=S_K) then BEGIN
      sKoenig:=von;
   END;
END;
{--------------------------------------------------------------------}
{- Bewertung eines schwarzen Bauern. Neben den Parameter muessen    -}
{- auch die Bauernkontrollen, Bauernlinien und Turmlinien richtig   -}
{- belegt sein.                                                     -}
{- Gibt die Bewertung aus der Sicht von Schwarz zurueck.            -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.sbauerbewertung(feld, reihe,  linie, entwickelt : integer) : integer;
var
   wert,j : Integer;
label fertig;
BEGIN
   reihe := (REIHE_8+REIHE_1)-reihe;  { Reihe umdrehen. Hoehere Reihe ist }
                                     { damit wie bei WEISS besser.       }
   if(MatSumme[Tiefe]>ENDSPIELMATERIAL) then BEGIN { Eroeffnung oder Mittelspiel }
      wert:=sBFeldWert[feld];
      { Entwicklung nicht abgeschlossen. Stosse Randbauern nicht vor }
      if((entwickelt<4)AND((linie>=F_LINIE)OR(linie<=B_LINIE)) AND
         (reihe>REIHE_3)) then BEGIN  { reihe wurde umgedreht. }
         wert := wert - 15;
      END;
   END
   else BEGIN  { Im Endspiel sind alle Linien gleich gut. Bringe Bauern nach }
           { vorne.                                                      }
      wert := reihe*4;
   END;
   {*
   *** Ist Bauer ein Isolani ?
   *** Randbauern brauchen nicht extra behandelt werden. Bauern[A_LINIE-1] ist
   *** linker Rand, Bauern[H_LINE+1] rechter Rand. Auf beiden Raendern steht
   *** kein Bauer.
   *}
   if((Bauern[linie-1].schwarz=0)AND(Bauern[linie+1].schwarz=0)) then BEGIN
      wert := wert - 12;  { Isolani }
      if(Bauern[linie].schwarz>1) then BEGIN { Isolierter Doppelbauer. }
         wert := wert - 12;
      END;
   END;
   if(Bauern[linie].schwarz>1) then BEGIN { Doppelbauer }
      wert := wert - 15;
   END;
   { Duo oder unterstuetzter Bauer bekommt Bonus }
   { Z.B. e5,d5 ist Duo, d6 unterstuetzt e5 }
   if((BKontrolle[feld].schwarz>0)OR(BKontrolle[feld-10].schwarz>0)) then BEGIN
      wert := wert + reihe ;
   END;
   if(Bauern[linie].weiss=0) then BEGIN { Halboffene Linie }
      { Rueckstaendier Bauer auf halboffener Linie }
      { Rueckstaendiger Bauer wird nicht von eigenen gestuetzt und  }
      { kann nicht nach vorne, weil gegnerische Bauern Feld vor ihm }
      { kontrollieren.                                              }
      if((BKontrolle[feld].schwarz=0)AND
         (BKontrolle[feld-10].weiss>BKontrolle[feld-10].schwarz)) then BEGIN
         wert := wert - 10;
         if((Tuerme[linie].weiss>0)) then BEGIN
            wert:=wert -8; { Turm kniet sich auf rueckstaendigen Bauer } 
         END;
      END
      else BEGIN
         {
         ** Bauer ist Freibauer, wenn auf halboffenen Linie und die
         ** weiteren Felder auf seiner Linie nicht von gegnerischen 
         ** Bauern kontrolliert werden.
         }
         j:=feld;
         while j>H2 do BEGIN
            if(BKontrolle[j].weiss>0) then BEGIN
               goto fertig;
            END;
            dec(j,10);
         END;
         { Freibauer gefunden. Im Endspiel ist Freibauer wichtiger als }
         { im Mittelspiel.                                             }
         if(MatSumme[Tiefe]<ENDSPIELMATERIAL) then BEGIN
            wert := wert + reihe*16; { Je weiter vorne, desto besser }
            if(Tuerme[linie].schwarz>0) then BEGIN { Turm unterstuetzt Freibauer }
               wert := wert + reihe*2;     { auf selben Linie            }
            END;
            if(Tuerme[linie].weiss>0) then BEGIN { gegnerischer Turm auf selben  }
               wert := wert - reihe*2;   { Linie.                        }
            END;
            if(MatSumme[Tiefe]=0) then BEGIN { Reines Bauernendspiel. Freibauer }
               wert := wert + reihe*8; { Besonders wertvoll.             }
            END;
            if((BKontrolle[feld].schwarz>0)OR
               (BKontrolle[feld-10].schwarz>0)) then BEGIN { Unterstuetzter Freibauer }
               wert := wert + reihe*4;
            END;
            { Freibauer von weisser Figur blockiert. Figur wird auch }
            { nicht von eigenen Bauern bedroht.                      }
            if((MMBrett[feld-10]<0)AND(BKontrolle[feld-10].schwarz=0)) then BEGIN
               wert:=wert - reihe*4;
            END;
         END
         else BEGIN { Freibauer im Mittelspiel }
            wert := wert + reihe * 8;
            if((BKontrolle[feld].schwarz>0)OR
               (BKontrolle[feld-10].schwarz>0)) then BEGIN { Unterstuetzter Freibauer }
               wert := wert + reihe*2;
            END;
         END;
      END;
   END;

fertig:
   sbauerbewertung:= wert;
END;
{--------------------------------------------------------------------}
{- Bewertung eines weissen Bauern.                                  -}
{- Erfolgt analog der Bewertung fuer schwarzen Bauern.              -}
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.wbauerbewertung(feld, reihe, linie, entwickelt : integer) : integer;
var
   wert,j : integer;
label  fertig;
BEGIN
   if(MatSumme[Tiefe]>ENDSPIELMATERIAL) then BEGIN
      wert:=wBFeldWert[feld];
      if((entwickelt<4)AND((linie>=F_LINIE)OR(linie<=B_LINIE))AND
         (reihe>REIHE_3)) then BEGIN
         wert := wert - 15;
      END;
   END
   else BEGIN
      wert := reihe*4;
   END;
   if((Bauern[linie-1].weiss=0)AND(Bauern[linie+1].weiss=0)) then BEGIN
      wert := wert - 12;  { Isolani }
      if(Bauern[linie].weiss>1) then BEGIN { Isolierter Doppelbauer. }
         wert := wert - 12;
      END;
   END;
   if(Bauern[linie].weiss>1) then BEGIN { Doppelbauer }
      wert := wert - 15;
   END;
   { Duo oder unterstuetzter Bauer bekommt Bonus }
   if((BKontrolle[feld].weiss>0)OR(BKontrolle[feld+10].weiss>0)) then BEGIN
      wert := wert + reihe;
   END;
   if(Bauern[linie].schwarz=0) then BEGIN { Halboffene Linie }
      { Rueckstandier Bauer auf halboffener Linie }
      if((BKontrolle[feld].weiss=0)AND
         (BKontrolle[feld+10].schwarz>BKontrolle[feld+10].weiss)) then BEGIN
         wert := wert - 10;
         if((Tuerme[linie].schwarz>0)) then BEGIN
            wert:=wert -8; { Turm kniet sich auf Bauer }
         END;
      END
      else BEGIN
         j:=feld;
         while j<A7 do BEGIN { Ueberpruefe Freibauer }
            if(BKontrolle[j].schwarz>0) then BEGIN
               goto fertig;
            END;
            dec(j,10);
         END;
         { Freibauer gefunden }
         if(MatSumme[Tiefe]<ENDSPIELMATERIAL) then BEGIN
            wert := wert + reihe*16;
            if(Tuerme[linie].weiss>0) then BEGIN { Turm unterstuetzt Freibauer }
               wert := wert + reihe*2;
            END;
            if(Tuerme[linie].schwarz>0) then BEGIN { gegnerischer Turm }
               wert := wert - reihe*2;
            END;
            if(MatSumme[Tiefe]=0) then BEGIN { Reines Bauernendspiel }
               wert := wert + reihe*8;
            END;
            if((BKontrolle[feld].weiss>0)OR
               (BKontrolle[feld+10].weiss>0)) then BEGIN { Unterstuetzter Freibauer }
               wert := wert + reihe*4;
            END;
            { Freibauer von schwarzer Figur blockiert. }
            if((MMBrett[feld+10]<0)AND(BKontrolle[feld+10].weiss=0)) then BEGIN
               wert:=wert - reihe*4;
            END;
         END
         else BEGIN { Freibauer im Mittelspiel }
            wert := wert + reihe * 8;
            if((BKontrolle[feld].weiss>0)OR
               (BKontrolle[feld+10].weiss>0)) then BEGIN { Unterstuetzter Freibauer }
               wert := wert + reihe*2;
            END;
         END;
      END;
   END;

fertig:
   wbauerbewertung:= wert;
END;

{--------------------------------------------------------------------}
{- Ueberprueft ob Spieler 'seite' das Feld 'feld' attackiert. Gibt   }
{- WAHR zurueck, wenn Feld von 'seite' attackiert wird, sonst FALSCH }
{- Methode: Stelle quasi auf 'feld' eine Superfigur, die in alle     }
{-          Richtungen ziehen kann. 'Schlaegt' diese Superfigur z.B. }
{-          einen Turm von 'seite', dann attackiert in Wirklichkeit  }
{-          der Turm das Feld.                                       } 
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.attackiert_feld(feld,  seite : integer ) : integer;
var
   i,richtung,nach,figur,gleiten : integer;
label weitergleiten, weiter, weiter1;
BEGIN
   { Teste zuerst Spezialfall Bauern. Haben dieselbe Richtung wie Laeufer,
   ** gleiten aber nicht.
   }
   if(seite=WEISS) then BEGIN
      { Muss umgekehrte Richtung als Bauer gehen. D5 wird von Bauer E4
      ** attackiert.
      }
      if((MMBrett[feld-9]=W_B)OR(MMBrett[feld-11]=W_B)) then BEGIN
         attackiert_feld:= WAHR;
         exit;
      END;
   END;
   if(seite=SCHWARZ) then BEGIN
      if((MMBrett[feld+9]=S_B)OR(MMBrett[feld+11]=S_B)) then BEGIN
         attackiert_feld:= WAHR;
         exit;
      END;
   END;
   {*
   *** Untersuche Springer
   *}
   i:=8;
   while i<16 do BEGIN { Springer Richtungen }
      nach:=feld+Offset[i];
      if((MMBrett[nach]=LEER)OR(MMBrett[nach]=RAND)) then BEGIN
         goto weiter;
      END;
      if(seite=WEISS) then BEGIN
         if(MMBrett[nach]=W_S) then BEGIN
            attackiert_feld:= WAHR;
            exit;
         END;
      END
      else if(MMBrett[nach]=S_S) then BEGIN
         attackiert_feld:= WAHR;
         exit;
      END;
      weiter:
      inc(i);
   END;
   {*
   *** Untersuche 'gleitende' Figuren und Koenig.
   *}
   for i:=0 to 7 do BEGIN
      nach:=feld;
      richtung:=Offset[i];
      gleiten:=0;
weitergleiten:
      gleiten:=gleiten+1;
      nach:=nach+richtung;
      if(MMBrett[nach]=LEER) then BEGIN
         goto weitergleiten;
      END;
      if(MMBrett[nach]=RAND) then BEGIN
         goto weiter1;  { Am Rand angelangt, naechste Richtung }
      END;
      { Auf Figur getroffen. Figur muss von Farbe 'seite' sein.
      ** Ausserdem muss die aktuelle Richtung eine moegliche Zugrichtung der
      ** Figur sein. Der Koenig darf weiters nur einen Schritt tun.
      }
      figur:=MMBrett[nach];
      if(seite=WEISS) then BEGIN
         if((figur>0)) then BEGIN  { Weisse Figur }
            if(figur=W_K) then BEGIN
               if(gleiten<=1) then BEGIN  { Koenig ist Kurzschrittler }
                  attackiert_feld:=WAHR;
                  exit;
               END;
            END
            else BEGIN { Bei gleitenden Figuren muss aktuelle Richtung eine
                   ** moegliche Zugrichtung der Figur sein.
                   }
               if((FigOffset[figur].anfang<=i)AND(FigOffset[figur].ende>=i)) then BEGIN
                  attackiert_feld:= WAHR;
                  exit;
               END;
            END;
         END;
      END
      else BEGIN  { fuer seite SCHWARZ }
         if(figur<0) then BEGIN
            if(figur=S_K) then BEGIN
               if(gleiten<=1) then BEGIN
                  attackiert_feld:= WAHR;
                  exit;
               END;
            END
            else BEGIN
               if((FigOffset[abs(figur)].anfang<=i)AND
                  (FigOffset[abs(figur)].ende>=i)) then BEGIN
                  attackiert_feld:= WAHR;
                  exit;
               END;
            END;
         END;
      END;
      weiter1:
   END;
   {*
   *** Alle Richtungen durchsucht, auf keine passende Figur gestossen.
   *** D.h. 'seite' attackiert das Feld nicht.
   *}
   attackiert_feld:= FALSCH;
END;
{--------------------------------------------------------------------}
{- Stellungsbewertung. Wert wird aus der Sicht von 'seite' zurueck   }
{- gegeben. Wenn Materialwert zu weit von alpha-beta Fenster         }
{- abweicht, wird nur Material bewertet.                             }
{- Falls 'seite' Matt setzt, wird MATTWERT-Tiefe zurueck gegeben.    }
{- Fall  'seite' im Schach steht, wird Variabel InSchach auf WAHR    }
{- gesetzt.                                                          }
{- ACHTUNG: Die Funktion nimmt sowohl fuer Schach/Matt als auch fuer }
{-          die Koenigsopposition an, dass 'seite' am Zug ist.       }
{--------------------------------------------------------------------}
FUNCTION
TMiniMax.bewerte_stellung( alpha, beta, seite : integer) : integer;
var
   wert,poswert,i,j,k,feld,wlaeufer,slaeufer : integer;
   bauernanzahl : integer;
   matsum,wtauf7,stauf2 : integer;
   wentwickelt,sentwickelt: integer;
BEGIN
   { Ueberpruefe zuerst ob Gegner Matt ist }
   { oder 'seite' im Schach steht          }
   if(seite=WEISS) then BEGIN
      if(attackiert_feld(sKoenig,WEISS)=WAHR) then BEGIN
         bewerte_stellung:= MATTWERT-Tiefe;
         exit;
      END;
      InSchach:=attackiert_feld(wKoenig,SCHWARZ); { Steht 'seite' im Schach? }
   END
   else BEGIN
      if(attackiert_feld(wKoenig,SCHWARZ)=WAHR) then BEGIN
         bewerte_stellung:= MATTWERT-Tiefe;
         exit;
      END;
      InSchach:=attackiert_feld(sKoenig,WEISS);
   END;
   { Positionelle Bewertungsfaktoren wiegen starkes Materialungleichgewicht }
   { nicht auf. Daher wird in diesem Fall auf Positionsbewertung verzichtet }
   { Ausnahme: Spaetes Endspiel. Freibauern haben hohe Bewertung.           }
   {           Leichtfigur ohne Bauern ist wirkungslos.                      }
   wert:=MatBilanz[Tiefe];
   matsum:=MatSumme[Tiefe];
   if(matsum>(2*MAT_L)) then BEGIN
      if((wert<alpha-MAX_POS)OR(wert>beta+MAX_POS)) then BEGIN
         bewerte_stellung:= wert;
         exit;
      END;
   END;
   {
   ** Initialisiere die Turm und Bauernlinien und die Bauernkontrollen.
   ** Das koennte man wesentlich schneller auch bei der Ausfuehrung (und
   ** der Zuruecknahme) von  Zuegen inkrementell berechnen. Allerdings ist
   ** diese inkrementelle Berechnung auch wegen der Spezialfaelle Rochade, 
   ** Enpassant und Umwandlung laestig und fehleranfaellig.
   ** Koennte auch in init_bewertung Figurenliste aufbauen und dann in
   ** der zweiten Runde diese Figurenliste durchgehen (und nicht mehr
   ** das gesamte Brett).
   ** Die schnellste Loesung besteht darin, auch diese Figurenliste
   ** inkrementell zu berechnen. Allerdings werden dann die Funktionen 
   ** 'zugausfuehren' und zugzuruecknehmen' noch komplizierter.
   ** Nach dem KISS (Keep It Simple und Stupid) Prinzip wurde in Minimax
   ** diese Loesung gewaehlt.
   }
   init_bewertung;
   poswert:=0;
   {*
   *** Werden fuer Bewertung des Laeuferpaares verwendet.
   *}
   slaeufer:=0;
   wlaeufer:=0;
   {*
   *** Zur Feststellung von ungenuegenden Material verwendet.
   *}
   bauernanzahl:=0;
   { Weise Tuerme auf 7/8-ter Reihe, schwarze auf 1/2-ter }
   wtauf7:=0;
   stauf2:=0;
   {*
   *** Stand der Entwicklung: Rochiert und Leichtfiguren entwickelt.
   *}
   wentwickelt:=Rochade[WEISS+1];
   sentwickelt:=Rochade[SCHWARZ+1];
   if(Bewegt[B1]>0) then BEGIN   { Springer auf B1 entwickelt? }
      wentwickelt:=wentwickelt+1;
   END;
   if(Bewegt[C1]>0) then BEGIN
      wentwickelt:=wentwickelt+1;
   END;
   if(Bewegt[F1]>0) then BEGIN
      wentwickelt:=wentwickelt+1;
   END;
   if(Bewegt[G1]>0) then BEGIN
      wentwickelt:=wentwickelt+1;
   END;
   if(Bewegt[B8]>0) then BEGIN   { Springer auf B8 entwickelt? }
      sentwickelt:=sentwickelt+1;
   END;
   if(Bewegt[C8]>0) then BEGIN
      sentwickelt:=sentwickelt+1;
   END;
   if(Bewegt[F8]>0) then BEGIN
      sentwickelt:=sentwickelt+1;
   END;
   if(Bewegt[G8]>0) then BEGIN
      sentwickelt:=sentwickelt+1;
   END;
   {
   ** Lese ganzes Brett und bewerte jede einzelne Figur 
   ** Es wird aus der Sicht von Weiss bewertet. Bei den schwarzen
   ** Figuren bedeutet eine positive Bewertung, dass dieses Merkmal
   ** fuer Schwarz schlecht ist.
   }
   for i:=REIHE_1 to REIHE_8 do BEGIN
      feld:=i*10+A_LINIE;
      for j:=A_LINIE to H_LINIE do BEGIN
         case MMBrett[feld] of
          S_K:
            if(matsum<ENDSPIELMATERIAL) then BEGIN { Endspielbewertung fuer Koenig }
               { Zentralisiere im Endspiel den Koenig }
               poswert:=poswert-ZentrumTab[feld];
            END
            else BEGIN
               { Noch nicht rochiert, aber Rochaderecht verloren }
               if((Rochade[SCHWARZ+1]=FALSCH)AND
                  ((Bewegt[E8]>0)OR((Bewegt[H8]>0)AND(Bewegt[A8]>0)))) then BEGIN
                  poswert := poswert + 35;
               END;
              { Koenig moeglichst nicht im Zentrum }
               poswert:=poswert+ (4*ZentrumTab[feld]);
               for k:=-1 to 1 do BEGIN
                  { Bonus fuer Bauernschirm vor Koenig }
                  if(MMBrett[feld-10+k]=S_B) then BEGIN
                     poswert:=poswert-15;
                  END;
                  { Bauernschirm 2 Reihen vor Koenig }
                  if(MMBrett[feld-20+k]=S_B) then BEGIN
                     poswert:=poswert-6;
                  END;
                  { Malus fuer halboffene Linie die von gegnerischen Turm
                  ** besetzt ist.
                  }
                  if((Bauern[j+k].weiss=0)AND(Tuerme[j+k].weiss>0)) then BEGIN
                     poswert:=poswert+12;
                  END;
               END;
            END;
          S_D:
            { Vermeide Damenausfluege zu Beginn des Spieles }
            if(sentwickelt<4) then BEGIN
               if(feld<A8) then BEGIN
                  poswert:=poswert+15;
               END;
            END
            else BEGIN { Falls Entwicklung abgeschlossen, plaziere Dame in der
                   ** Naehe des gegn. Koenigs.
                   ** Erster Ausdruck ist Reihendifferenz, zweiter 
                   ** Spaltendifferenz !!!
                   }
               poswert:=poswert+ 2*(abs((wKoenig div 10)-(feld div 10)) +
                                   abs((wKoenig mod 10)-(feld mod 10)));
            END;
          S_S: { Schwarzer Springer }
            poswert:=poswert-(ZentrumTab[feld] div 2); { Zentralisiere Springer }
          S_L: BEGIN
              { Laeufer soll nicht schwarze d7/e7 Bauern verstellen         }
              { Laeufer wird auch im Zuggenerator durch Variable Mobilitaet }
              { bewertet.                                                   }
                 if(((feld=D6)OR(feld=E6))AND(MMBrett[feld+10]=S_B)) then BEGIN
                    poswert:=poswert+20;
                 END;
                 slaeufer:=slaeufer+1;  { Anzahl der Laeufer fuer Laeuferpaar }
               END;
           S_T:  { Turm geht auch bei Koenigsbewertung ein }
              BEGIN
                if(feld<=H2) then BEGIN
                   stauf2:=stauf2+1; { Schwarzer Turm ist auf Reihe 1 oder 2 }
                                { eingebrochen }
                END;
                { Tuerme von a- und h-Linien ins Zentrum bringen }
                if((j>=C_LINIE)AND(j<=E_LINIE)) then BEGIN
                   poswert:=poswert-4;
                END;
                { Tuerme auf halboffene und offene Linien }
                if(Bauern[j].weiss=0) then BEGIN
                   poswert:=poswert-8; { Turm auf halboffener Linie }
                   if(Bauern[j].schwarz=0) then BEGIN { auf offener Linie }
                      poswert:=poswert-5;
                   END;
                 END;
               END;
          S_B: { Bauernbewertung ist relative komplex. Wird daher }
                   { in eigener Routine erledigt.                     }
               BEGIN
                 poswert:=poswert-sbauerbewertung(feld,i,j,sentwickelt);
                 bauernanzahl:=bauernanzahl+1;
               END;
          LEER:BEGIN END;
          W_B:  { Weisse Bewertung ist analog der Schwarzen }
               BEGIN
                 poswert:=poswert+wbauerbewertung(feld,i,j,wentwickelt);
                 bauernanzahl:=bauernanzahl+1;
               END;
          W_T: BEGIN
                 if(feld>=A7) then BEGIN
                    wtauf7:=wtauf7+1; { weisser Turm auf 7-ter oder 8-ter Reihe }
                 END;
                 { Tuerme von a- und h-Linien ins Zentrum bringen }
                 if((j>=C_LINIE)AND(j<=E_LINIE)) then BEGIN
                    poswert:=poswert+4;
                 END;
                 { Tuerme auf halboffene und offene Linien bringen }
                 if(Bauern[j].schwarz=0) then BEGIN
                    poswert:=poswert+8; { Turm auf halboffener Linie }
                    if(Bauern[j].weiss=0) then BEGIN { auf offener Linie }
                       poswert:=poswert+5;
                    END;
                 END;
               END;
          W_L:
            { Laeufer soll nicht Bauern auf D3/E3 blockieren. }
               BEGIN
                 if(((feld=D3)OR(feld=E3))AND(MMBrett[feld-10]=W_B)) then BEGIN
                    poswert:=poswert-20;
                 END;
                 wlaeufer:=wlaeufer+1;
               END;
          W_S:
               poswert:=poswert+(ZentrumTab[feld] div 2);
          W_D:
            { Vermeide Damenausfluege zu Beginn des Spieles }
            if(wentwickelt<4) then BEGIN
               if(feld>H1) then BEGIN
                  poswert:=poswert-15;
               END;
            END
            else BEGIN { Plaziere Dame in der Naehe des gegn. Koenigs }
                   { Erster Ausdruck ist Reihendifferenz, zweiter }
                   { Spaltendifferenz }
               poswert:=poswert- 2*(abs((sKoenig div 10)-(feld div 10)) +
                                   abs((sKoenig mod 10)-(feld mod 10)));
            END;
          W_K:
            if(matsum<ENDSPIELMATERIAL) then BEGIN { Endspielbewertung fuer Koenig }
               { Zentralisiere im Endspiel den Koenig }
               poswert:=poswert+ZentrumTab[feld];
               { Nah-Opposition der Koenige }
               if((abs(feld-sKoenig)=20)OR(abs(feld-sKoenig)=2)) then BEGIN
                  k:=10;
                  if(matsum=0) then BEGIN  { Opposiotion im Bauernendspiel }
                     k:=30;
                  END;
                  if(Farbe=WEISS) then BEGIN
                     poswert := poswert - k;
                  END
                  else BEGIN
                     poswert := poswert + k;
                  END;
               END;
            END
            else BEGIN
               { Noch nicht rochiert, aber Rochaderecht verloren }
               if((Rochade[WEISS+1]=FALSCH)AND
                  ((Bewegt[E1]>0)OR((Bewegt[H1]>0)AND(Bewegt[A1]>0)))) then BEGIN
                  poswert := poswert - 35;
               END;
              { Koenig moeglichst nicht im Zentrum }
               poswert:=poswert- (4*ZentrumTab[feld]);
               for k:=-1 to 1 do BEGIN
                  { Bonus fuer Bauernschirm vor Koenig }
                  if(MMBrett[feld+10+k]=W_B) then BEGIN
                     poswert:=poswert+15;
                  END;
                  { Bauernschirm 2 Reihen vor Koenig }
                  if(MMBrett[feld+20+k]=W_B) then BEGIN
                     poswert:=poswert+6;
                  END;
                  { Malus fuer Halboffene Linie die von gegnerischen Turm
                  ** besetzt ist.
                  }
                  if((Bauern[j+k].schwarz=0)AND(Tuerme[j+k].schwarz>0)) then BEGIN
                     poswert:=poswert-12;
                  END;
               END;
            END;
         END;
         feld:=feld+1
      END;
   END;
   { Keine Bauern mehr am Brett und ungenuegendes Material }
   { Erkennt alle elementaren Remise Situationen.          }
   { (K,K), (KL,K) (KS,K) (KSS,K) (KL,KL) (KS,KL)          }
   { nicht aber (KSS,KL). Ist das immer Remise?            }
   if(bauernanzahl=0) then BEGIN
      if((matsum<=MAT_L)OR(matsum=2*MAT_S)OR
         ((abs(MatBilanz[Tiefe])<100)AND(matsum<=(2*MAT_L)))) then BEGIN
         bewerte_stellung:= 0;
         exit;
      END;
   END;
   if(wlaeufer>=2) then BEGIN  { Laeuferpaar Bonus fuer Weiss }
      poswert:=poswert+15;
   END;
   if(slaeufer>=2) then BEGIN  { Laeuferpaar Bonus fuer Schwarz }
      poswert:=poswert-15;
   END;
   { wTuerme auf 7/8-ter Reihe und schwarze Koenig auch noch auf diesen }
   { Reihen }
   if((wtauf7>0)AND(sKoenig>=A7)) then BEGIN
      poswert:=poswert+10;
      if(wtauf7>1) then BEGIN  { Verdoppelt ist extra gefaehrlich }
         poswert:=poswert+25;
      END;
   END;
   { sTuerme auf 1/2-ter Reihe und weisser Koenig auch noch auf diesen }
   { Reihen }
   if((stauf2>0)AND(wKoenig<=H2)) then BEGIN
      poswert:=poswert-10;
      if(stauf2>1) then BEGIN
         poswert:=poswert-25;
      END;
   END;
   if(seite=SCHWARZ) then BEGIN  { Bewertung erfolgte aus der Sicht von Weiss }
      poswert:= -poswert; { Drehe daher fuer Schwarz wert um           }
   END;
   { Beruecksichtige die Mobilitaet der Lauefer und Tuerme
   ** Wird von Zuggenerator berechnet. Mobilitaet[Tiefe] ist die
   ** Mobilitaet der anderen Farbe, Mobilitaet[Tiefe-1] jene der
   ** 'seite' (bevor die anderen Farbe einen Zug ausgefuehrt hat).
   }
   if(Tiefe>=1) then BEGIN
      poswert:=poswert-((Mobilitaet[Tiefe]-Mobilitaet[Tiefe-1]) div 16);
   END;
   bewerte_stellung:= wert+poswert;
END;

{--------------------------------------------------------------------}
{- Trage normalen Zug in Zugstapel ein.                             -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.notiere_zug( von, nach : integer);
BEGIN
   { 
   ** Erhoehe die Mobilitaet fuer Laeufer und Turm.
   ** Mobilitaet im Zentrum wird hoeher gewichtet
   ** als Mobilitaet am Rand.
   }
   if(Farbe=WEISS) then BEGIN
      if((MMBrett[von]=W_L)OR(MMBrett[von]=W_T)) then BEGIN
         Mobilitaet[Tiefe]:=Mobilitaet[Tiefe] + ZentrumTab[nach];
      END;
   END
   else BEGIN
      if((MMBrett[von]=S_L)OR(MMBrett[von]=S_T)) then BEGIN
         Mobilitaet[Tiefe]:=Mobilitaet[Tiefe]+ZentrumTab[nach];
      END;
   END;
   {*
   *** Bewerte Zug fuer Zugsortierung. Bonus fuer Hauptvariante oder 'Killer'
   *}
   if((HVar[0][Tiefe].von=von)AND(HVar[0][Tiefe].nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=HVARBONUS;
   END
   else if((KillerTab[Tiefe].killer1.von=von)AND
      (KillerTab[Tiefe].killer1.nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=KILLER1BONUS;
   END
   else if((KillerTab[Tiefe].killer2.von=von)AND
      (KillerTab[Tiefe].killer2.nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=KILLER2BONUS;
   END
   else  BEGIN
      ZugStapel^[Index].wert:=LEER;
   END;
   ZugStapel^[Index].von:=von;
   ZugStapel^[Index].nach:=nach;
   ZugStapel^[Index].geschl_fig:=LEER;
   ZugStapel^[Index].umwandl_fig:=LEER;
   ZugStapel^[Index].rochade_nr:=KEINEROCHADE;
   ZugStapel^[Index].ep_feld:=ILLEGAL;
   if(Index<ZSTAPEL_DIM) then BEGIN { Verhindere Ueberlauf des Zugstapels }
      Index:=Index+1;
   END
   else BEGIN
      sllines.add('ERROR: Zugstapel Ueberlauf');
      if assigned(OnTrace) then OnTrace(self, slLines.Text);
      status:=fehler;
      PStop;
   END;
END;
{--------------------------------------------------------------------}
{- Trage SchlagZug in Zugstapel ein.                                -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.notiere_schlagzug( von, nach : integer);
var
   figwert : integer;
BEGIN
   { Koenig kann nicht geschlagen werden }
   if((MMBrett[nach]=W_K)OR(MMBrett[nach]=S_K)) then BEGIN
      exit;
   END;
   figwert:=FigMaterial[abs(MMBrett[nach])];
   ZugStapel^[Index].von:=von;
   ZugStapel^[Index].nach:=nach;
   ZugStapel^[Index].geschl_fig:=abs(MMBrett[nach]);
   {*
   *** Schlage mit moeglichst kleiner Figur moeglichst wertvolle
   *}
   ZugStapel^[Index].wert:=figwert-(FigMaterial[abs(MMBrett[von])] div 8);
   {*
   *** Extra Bonus fuer das Schlagen der zuletzt gezogenen Figur 
   *}
   if((Tiefe>0)AND(nach=NachFeld[Tiefe-1])) then BEGIN
      ZugStapel^[Index].wert:=ZugStapel^[Index].wert+300;
   END;
   {*
   *** Bonus fuer Hauptvariante Zuege und 'Killer'
   *}
   if((HVar[0][Tiefe].von=von)AND(HVar[0][Tiefe].nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=ZugStapel^[Index].wert+HVARBONUS;
   END
   else if((KillerTab[Tiefe].killer1.von=von)AND
      (KillerTab[Tiefe].killer1.nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=ZugStapel^[Index].wert+KILLER1BONUS;
   END
   else if((KillerTab[Tiefe].killer2.von=von)AND
      (KillerTab[Tiefe].killer2.nach=nach)) then BEGIN
      ZugStapel^[Index].wert:=ZugStapel^[Index].wert+KILLER2BONUS;
   END;
   ZugStapel^[Index].umwandl_fig:=LEER;
   ZugStapel^[Index].rochade_nr:=KEINEROCHADE;
   ZugStapel^[Index].ep_feld:=ILLEGAL;
   if(Index<ZSTAPEL_DIM) then BEGIN { Verhindere Ueberlauf des Zugstapels }
      Index:=Index+1;
   END
   else BEGIN
      sllines.add('ERROR: Zugstapel Ueberlauf');
      if assigned(OnTrace) then OnTrace(self, slLines.Text);
      Status:=fehler;
      PStop;
   END;
END;
{--------------------------------------------------------------------}
{- Trage Enpassant-Zug in Zugstapel ein.                            -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.notiere_epzug( von,nach, epfeld : integer);
BEGIN
   { Koenig kann nicht geschlagen werden }
   if((MMBrett[nach]=W_K)OR(MMBrett[nach]=S_K)) then BEGIN
      exit;
   END;
   ZugStapel^[Index].von:=von;
   ZugStapel^[Index].nach:=nach;
   ZugStapel^[Index].geschl_fig:=W_B;
   ZugStapel^[Index].umwandl_fig:=LEER;
   ZugStapel^[Index].rochade_nr:=KEINEROCHADE;
   ZugStapel^[Index].ep_feld:=epfeld;
   ZugStapel^[Index].wert:=MAT_B;
   if(Index<ZSTAPEL_DIM) then BEGIN { Verhindere Ueberlauf des Zugstapels }
      Index:=Index+1;
   END
   else BEGIN
      sllines.add('ERROR: Zugstapel Ueberlauf');
      if assigned(OnTrace) then OnTrace(self, slLines.Text);
      Status:=fehler;
      PStop;
   END;
END;
{--------------------------------------------------------------------}
{- Erzeuge alle moeglichen Bauerumwandlungen.                       -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.notiere_umwandlung(von, nach : integer);
var
   i : integer;
BEGIN
   if(MMBrett[nach]=LEER) then BEGIN
      i:=W_D;
      while i>W_B do BEGIN { Reihenfolge Dame,Springer,Laeufer,Turm }
         notiere_zug(von,nach);
         ZugStapel^[Index-1].umwandl_fig:=i;
         dec(i);
      END;
   END
   else BEGIN
      i:=W_D;
      while i>W_B do BEGIN
         notiere_schlagzug(von,nach);
         ZugStapel^[Index-1].umwandl_fig:=i;
         dec(i);
      END;
   END;
END;

{--------------------------------------------------------------------}
{- Generiert Zuege und legt sie auf ZugStapel.                       }
{- Falls 'allezuege' groesser 0 ist, werden alle pseudo legalen      }
{- Zuege erzeugt, sonst alle pseudo legalen Schlagzuege,             }
{- Umwandlungen, Enpassant und die Rochade.                          }
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.generiere_zuege(allezuege : integer);
var
   von,nach,figur : integer;
   i, j, langschritt,richtung : integer;
   epfeld : integer;
   moeglich : Boolean;
label weiter, weiter1;
label weitergleiten;
BEGIN
   Index:=StapelGrenze[Tiefe]; { Anfang der Zugliste auf Tiefe T }
   Mobilitaet[Tiefe]:=0;
   {
   ** Durchsuche das Brett nach Figuren
   }
   for von:=A1 to H8 do BEGIN
      figur:=MMBrett[von];
      if((figur=LEER)OR(figur=RAND)) then BEGIN  { Leer- und Randfeld machen }
         goto weiter;                        { keine Zuege }
      END;
      { Figur muss auch von richtiger Farbe sein }
      if(((Farbe=WEISS)AND(figur<0))OR((Farbe=SCHWARZ)AND(figur>0))) then BEGIN
         goto weiter;
      END;
      figur:=abs(figur);   { Art der Figur. Fuer Zugrichtung ist - bis
                          ** auf Bauern - Farbe egal
                          }
      if(figur=W_B) then BEGIN { Bauern Zuege }
         if(Farbe=WEISS) then BEGIN
            if(MMBrett[von+10]=LEER) then BEGIN
               if(von>=A7) then BEGIN
                  notiere_umwandlung(von,von+10);
               END
               else if(allezuege>0) then BEGIN
                  notiere_zug(von,von+10);
                  { Doppelschritt moeglich? }
                  if((von<=H2)AND(MMBrett[von+20]=LEER)) then BEGIN
                     notiere_zug(von,von+20);
                     { Notiere Zug hat Index schon erhoeht }
                     ZugStapel^[Index-1].ep_feld:=von+10;
                  END;
               END;
            END;
            if(MMBrett[von+11]<0) then BEGIN { Bauer kann Schwarze Figur schlagen }
               if(von>=A7) then BEGIN
                  notiere_umwandlung(von,von+11);
               END
               else BEGIN
                  notiere_schlagzug(von,von+11);
               END;
            END;
            if(MMBrett[von+9]<0) then BEGIN  { Ebenso in anderer Schlagrichtung }
               if(von>=A7) then BEGIN
                  notiere_umwandlung(von,von+9);
               END
               else BEGIN
                  notiere_schlagzug(von,von+9);
               END;

            END;
         END
         else if(Farbe=SCHWARZ) then BEGIN   { Selbiges fuer schwarze Bauern }
            if(MMBrett[von-10]=LEER) then BEGIN
               if(von<=H2) then BEGIN
                  notiere_umwandlung(von,von-10);
               END
               else if(allezuege>0) then BEGIN
                  notiere_zug(von,von-10);
                  { Doppelschritt moeglich? }
                  if((von>=A7)AND(MMBrett[von-20]=LEER)) then BEGIN
                     notiere_zug(von,von-20);
                     { Notiere Zug hat Index schon erhoeht }
                     ZugStapel^[Index-1].ep_feld:=von-10;
                  END;
               END;
            END;
            { Bei schwarzen Bauern muss man auch nach Rand abfragen }
            { bei weissen nicht, da der Rand > 0 ist.               }
            if((MMBrett[von-11]>0)AND(MMBrett[von-11]<>RAND)) then BEGIN
               if(von<=H2) then BEGIN
                  notiere_umwandlung(von,von-11);
               END
               else BEGIN
                  notiere_schlagzug(von,von-11);
               END;
            END;
            if((MMBrett[von-9]>0)AND(MMBrett[von-9]<>RAND)) then BEGIN
               if(von<=H2) then BEGIN
                  notiere_umwandlung(von,von-9);
               END
               else BEGIN
                  notiere_schlagzug(von,von-9);
               END;
            END;
         END;
         goto weiter;   { Untersuche naechstes Feld }
      END;
      {*
      *** Zuege fuer alle anderen Figuren werden ueber Zugoffset
      *** berechnet.
      *}
      langschritt:=FigOffset[figur].langschritt;
      i:=FigOffset[figur].anfang;
      while i<=FigOffset[figur].ende do BEGIN
         richtung:=Offset[i];
         nach:=von;
weitergleiten:
         nach:=nach+richtung;
         if(MMBrett[nach]=LEER) then BEGIN
            if(allezuege>0) then BEGIN
               notiere_zug(von,nach);
            END;
            if(langschritt=WAHR) then BEGIN { Laeufer, Turm und Dame }
               goto weitergleiten;
            END
            else BEGIN  { Springer und Koenig }
               goto weiter1;
            END;
         END;
         if(MMBrett[nach]=RAND) then BEGIN { Auf Randfeld getroffen, in anderer }
            goto weiter1;            { Richtung weitersuchen.             }
         END;
         { Auf Figur gestossen. Muss von richtiger Farbe sein. }
         if(((Farbe=WEISS)AND(MMBrett[nach]<0))OR(
             (Farbe=SCHWARZ)AND(MMBrett[nach]>0))) then BEGIN
            notiere_schlagzug(von,nach);
         END;
         weiter1:
         inc(i);
      END;
      weiter:
   END;
   {*
   *** Enpassant Zug
   *}
   if(Ep_Feld[Tiefe]<>ILLEGAL) then BEGIN
      epfeld:=Ep_Feld[Tiefe];
      if(Farbe=WEISS) then BEGIN
         if(MMBrett[epfeld-9]=W_B) then BEGIN
            notiere_epzug(epfeld-9,epfeld,epfeld-10);
         END;
         if(MMBrett[epfeld-11]=W_B) then BEGIN
            notiere_epzug(epfeld-11,epfeld,epfeld-10);
         END;
      END
      else BEGIN
         if(MMBrett[epfeld+9]=S_B) then BEGIN
            notiere_epzug(epfeld+9,epfeld,epfeld+10);
         END;
         if(MMBrett[epfeld+11]=S_B) then BEGIN
            notiere_epzug(epfeld+11,epfeld,epfeld+10);
         END;
      END;
   END;
   { Rochade wird auch in Ruhesuche ausgefuehrt, weil sie die Bewertung }
   { stark veraendert. (Ob das Sinnvoll ist, darueber scheiden sich     }
   { auch bei den Spitzenprogrammen die Geister).                       }
   {*
   *** Berechne Rochade
   *}
   if(Farbe=WEISS) then BEGIN
      if assigned(FRC) then with FRC^ do begin
        moeglich := false;
        if((wKoenig=FischerWhiteKing)AND(Bewegt[FischerWhiteKing]=0)) and
        (attackiert_feld(FischerWhiteKing,SCHWARZ)=FALSCH) then BEGIN
          if((MMBrett[FischerRightWhiteRook]=W_T)AND(Bewegt[FischerRightWhiteRook]=0))AND
          ((MMBrett[f1] = LEER) or (f1 = FischerRightWhiteRook) or (f1 = FischerWhiteKing)) AND
          ((MMBrett[g1] = LEER) or (g1 = FischerRightWhiteRook) or (g1 = FischerWhiteKing)) then begin
            moeglich := true;
            j := FischerWhiteKing + 1;
            while (j < FischerRightWhiteRook) and (j < H1) do begin
              if not (
                ((MMBrett[j]=LEER) or (j = FischerRightWhiteRook)) and
                (attackiert_feld(j,SCHWARZ)=FALSCH)
              )
              then begin
                moeglich := false;
                break;
              end;
              inc(j);
            end;
          end;
          if moeglich then begin
            notiere_zug(FischerWhiteKing,G1); { Notiere Koenigszug }
            ZugStapel^[Index-1].rochade_nr:=KURZEFISCHERROCHADE;
          end;
          if((MMBrett[FischerLeftWhiteRook]=W_T)AND(Bewegt[FischerLeftWhiteRook]=0))AND
          ((MMBrett[d1] = LEER) or (d1 = FischerLeftWhiteRook) or (d1 = FischerWhiteKing)) AND
          ((MMBrett[c1] = LEER) or (c1 = FischerLeftWhiteRook) or (c1 = FischerWhiteKing)) then begin
            moeglich := true;
            if FischerWhiteKing = B1 then j := C1
            else j := FischerWhiteKing - 1;
            while (j > FischerLeftWhiteRook) and (j > B1) do begin
              if not (
                ((MMBrett[j]=LEER) or (j = FischerLeftWhiteRook)) and
                (attackiert_feld(j,SCHWARZ)=FALSCH)
              )
              then begin
                moeglich := false;
                break;
              end;
              dec(j);
            end;
            if moeglich then begin
              notiere_zug(FischerWhiteKing,C1); { Notiere Koenigszug }
              ZugStapel^[Index-1].rochade_nr:=LANGEFISCHERROCHADE;
            end;
          end;
        end;
      end
      else begin
        if((wKoenig=E1)AND(Bewegt[E1]=0)) then BEGIN
          if((MMBrett[H1]=W_T)AND(Bewegt[H1]=0)AND
            (MMBrett[F1]=LEER)AND(MMBrett[G1]=LEER)AND
            (attackiert_feld(E1,SCHWARZ)=FALSCH)AND
            (attackiert_feld(F1,SCHWARZ)=FALSCH)AND
            (attackiert_feld(G1,SCHWARZ)=FALSCH)) then BEGIN
            notiere_zug(E1,G1); { Notiere Koenigszug }
            ZugStapel^[Index-1].rochade_nr:=KURZEROCHADE;
          END;
          if((MMBrett[A1]=W_T)AND(Bewegt[A1]=0)AND
            (MMBrett[D1]=LEER)AND(MMBrett[C1]=LEER)AND
            (MMBrett[B1]=LEER)AND
            (attackiert_feld(E1,SCHWARZ)=FALSCH)AND
            (attackiert_feld(D1,SCHWARZ)=FALSCH)AND
            (attackiert_feld(C1,SCHWARZ)=FALSCH)) then BEGIN
            notiere_zug(E1,C1); { Notiere Koenigszug }
            { Notiere Art der Rochade }
            ZugStapel^[Index-1].rochade_nr:=LANGEROCHADE;
          END;
        END;
      end;
   END
   else BEGIN { Schwarz am Zug }
      if assigned(FRC) then with FRC^ do begin
        moeglich := false;
        if((sKoenig=FischerBlackKing)AND(Bewegt[FischerBlackKing]=0))AND
        (attackiert_feld(FischerBlackKing,WEISS)=FALSCH) then BEGIN
          if((MMBrett[FischerRightBlackRook]=S_T)AND(Bewegt[FischerRightBlackRook]=0))AND
          ((MMBrett[f8] = LEER) or (f8 = FischerRightBlackRook) or (f8 = FischerBlackKing)) AND
          ((MMBrett[g8] = LEER) or (g8 = FischerRightBlackRook) or (g8 = FischerBlackKing)) then begin
            moeglich := true;
            j := FischerBlackKing + 1;
            while (j < FischerRightBlackRook) and (j < H8) do begin
              if not (
                ((MMBrett[j]=LEER) or (j = FischerRightBlackRook)) and
                (attackiert_feld(j,WEISS)=FALSCH)
              )
              then begin
                moeglich := false;
                break;
              end;
              inc(j);
            end;
            if moeglich then begin
              notiere_zug(FischerBlackKing,G8); { Notiere Koenigszug }
              ZugStapel^[Index-1].rochade_nr:=KURZEFISCHERROCHADE;
            end;
          end;
          if((MMBrett[FischerLeftBlackRook]=S_T)AND(Bewegt[FischerLeftBlackRook]=0))AND
          ((MMBrett[d8] = LEER) or (d8 = FischerLeftBlackRook) or (d8 = FischerBlackKing)) and
          ((MMBrett[c8] = LEER) or (c8 = FischerLeftBlackRook) or (c8 = FischerBlackKing)) then begin
            moeglich := true;
            if FischerBlackKing = B8 then j := C8
            else j := FischerBlackKing - 1;
            while (j > FischerLeftBlackRook) and (j > B8) do begin
              if not (
                ((MMBrett[j]=LEER) or (j = FischerLeftBlackRook)) and
                (attackiert_feld(j,WEISS)=FALSCH)
              )
              then begin
                moeglich := false;
                break;
              end;
              dec(j);
            end;
            if moeglich then begin
              notiere_zug(FischerBlackKing,C8); { Notiere Koenigszug }
              ZugStapel^[Index-1].rochade_nr:=LANGEFISCHERROCHADE;
            end;
          end;
        end;
      end
      else begin
        if((sKoenig=E8)AND(Bewegt[E8]=0)) then BEGIN
          if((MMBrett[H8]=S_T)AND(Bewegt[H8]=0)AND
            (MMBrett[F8]=LEER)AND(MMBrett[G8]=LEER)AND
            (attackiert_feld(E8,WEISS)=FALSCH)AND
            (attackiert_feld(F8,WEISS)=FALSCH)AND
            (attackiert_feld(G8,WEISS)=FALSCH)) then BEGIN
            notiere_zug(E8,G8); { Notiere Koenigszug }
            ZugStapel^[Index-1].rochade_nr:=KURZEROCHADE;
          END;
          if((MMBrett[A8]=S_T)AND(Bewegt[A8]=0)AND
            (MMBrett[D8]=LEER)AND(MMBrett[C8]=LEER)AND
            (MMBrett[B8]=LEER)AND
            (attackiert_feld(E8,WEISS)=FALSCH)AND
            (attackiert_feld(D8,WEISS)=FALSCH)AND
            (attackiert_feld(C8,WEISS)=FALSCH)) then BEGIN
            notiere_zug(E8,C8); { Notiere Koenigszug }
            { Notiere Art der Rochade }
            ZugStapel^[Index-1].rochade_nr:=LANGEROCHADE;
          END;
        END;
      end;
   END;
   StapelGrenze[Tiefe+1]:=Index;         { Merke Ende der Zugliste }
END;

{--------------------------------------------------------------------}
{- Berechnet die Bauernkontrolle und die Linien auf denen Bauern und }
{- Tuerme stehen. Wird von der Bewertungsfunktion zu Beginn          }
{- aufgerufen.                                                       }
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.init_bewertung;
var
   i : integer;
label weiter;
BEGIN
   {*
   *** Loesche Bauernkontrolle (Funktion memcpy ist natuerlich effektiver)
   **}
   for i:=A1 to H8 do BEGIN
      BKontrolle[i].weiss:=0;
      BKontrolle[i].schwarz:=0;
   END;
   {*
   *** Raender mit initialisieren. Damit braucht man bei Randlinien nicht
   *** extra abfragen.
   *}
   for i:=0 to H_LINIE+1 do BEGIN
      Bauern[i].weiss:=0;
      Bauern[i].schwarz:=0;
      Tuerme[i].weiss:=0;
      Tuerme[i].schwarz:=0;
   END;
   for i:=A1 to H8 do BEGIN
      if((MMBrett[i]=LEER)OR(MMBrett[i]=RAND)) then BEGIN
         goto weiter;
      END;
      case (MMBrett[i]) of
       W_B:BEGIN
             BKontrolle[i+9].weiss:=BKontrolle[i+9].weiss+1;
             BKontrolle[i+11].weiss:=BKontrolle[i+11].weiss+1;
             Bauern[i mod 10].weiss:=Bauern[i mod 10].weiss+1;
           END;
       S_B:BEGIN
             BKontrolle[i-9].schwarz:=BKontrolle[i-9].schwarz+1;
             BKontrolle[i-11].schwarz:=BKontrolle[i-11].schwarz+1;
             Bauern[i mod 10].schwarz:=Bauern[i mod 10].schwarz+1;
           END;
       S_T:BEGIN
             Tuerme[i mod 10].schwarz:=Tuerme[i mod 10].schwarz+1;
           END;
       W_T:BEGIN
             Tuerme[i mod 10].weiss:=Tuerme[i mod 10].weiss+1;
           END;
      END;
      weiter:
   END;
END;
{--------------------------------------------------------------------}
{- Berechnet den naechsten Zug des Computers.                       -}
{- Die Suche wird iterativ bis zur Solltiefe vertieft. Die Suche    -}
{- verwendet 'aspiration' Alpha-Beta.                               -}
{- Der Suchprozess kann mit einem beliebigen Tastendruck            -}
{- abgebrochen werden.                                              -}
{- Falls die Suche nicht abgebrochen wurde und kein Matt/Patt       -}
{- vorliegt, wird der beste Zug ausgefuehrt.                        -}
{--------------------------------------------------------------------}
PROCEDURE
TMiniMax.computer_zug;
var
   distanz,alpha,beta,i,wert,besterwert,j : integer;
   schach : integer;
   tmp : zug_t;
Label abgebrochen;
BEGIN
   init_spielbaum;
   { Bewerte Ausgangsstellung. Beende Suche falls Gegner schon Matt ist }
   wert:=bewerte_stellung(-MATTWERT,MATTWERT,Farbe);
   if(wert=MATTWERT) then BEGIN
      exit;
   END;
   { Speichere 'Schachstatus'. Wird fuer Patterkennung am Ende der Suche
   ** benoetigt.
   }
   schach:=InSchach;
   KnotenAnzahl:=0;
   {*
   *** Generiere alle pseudo-legalen Zuegen.
   *}
   generiere_zuege(1);
   {*
   *** Man koennte/sollte hier alle illegalen Zuege aus dem Zugstapel ent
   *** fernen und nur mit den legalen Zuegen weiter rechnen.
   *** (Hat aber nur optische Effekte, da Suche nach der Ausfuehrung des
   ***  illegalen Zuges sowieso sofort abbricht).
   *}

   {* Iteratives Vertiefen. distanz ist die Anzahl der Halbzuege bis zum
   *** Horizont. Ist insofern nicht gleich der Tiefe, da im Laufe des
   *** Suchprozesses die Distanz erhoeht werden kann (z.B. bei Schachs ).
   *}
   distanz:=1;
   while distanz<=SollTiefe do BEGIN
      if(distanz=1) then BEGIN   { Auf Tiefe 1 wird mit offenen Fenster gerechnet }
         alpha:=-MATTWERT; { Wir haben noch keinen guten Schaetzwert fuer  }
         beta:=MATTWERT;   { die Stellung.                                 }
      END
      else BEGIN              { Auf den hoehren Stufen sollte sich Resultat }
         beta:=alpha+100;  { nicht wesentlich vom Resultat der vorher-   }
         alpha:=alpha-100; { gehenden Tiefe unterscheiden.               }
      END;
      {
      ** Bei Schlagzuegen und Schachs wird Suche erweitert. Diese Variable
      ** begrenzt die Erweiterungen.
      }
      MaxErweiterung:=distanz+3;

      {*
      *** Berechne den Wert jedes Zuges 
      *}
      i:=0;
      while i<StapelGrenze[1] do BEGIN

         { Initialisiere die Hauptvariante und zeige den gerade }
         { untersuchten Zug an.                                 }
         HVar[Tiefe][Tiefe].von:=0;

         { Zug ausfuehren, wert berechnen, Zug zuruecknehmen. }
         zug_ausfuehren(i);
         wert:=-alpha_beta(-beta,-alpha,distanz-1);
         zug_zuruecknehmen(i);

         if stop then goto abgebrochen;
         if(i=0) then BEGIN { War es der erste Zug (der bisher beste)?         }
                    { Fuer diesen Zug wird ein exakter Wert benoetigt. }
            if(wert<alpha) then BEGIN
               { Suche des bisher besten Zuges faellt nach unten aus dem  }
               { Fenster( Programm erkennt Unglueck). Neuerliche Suche    }
               { mit nach unten geoeffneten Fenster notwendig.            }
               alpha:=-MATTWERT;
               beta:=wert;
               zug_ausfuehren(i);
               wert:=-alpha_beta(-beta,-alpha,distanz-1);
               zug_zuruecknehmen(i);
            END
            else if(wert>=beta) then BEGIN { Faellt nach oben }
               alpha:=wert;
               beta:=MATTWERT;
               zug_ausfuehren(i);
               wert:=-alpha_beta(-beta,-alpha,distanz-1);
               zug_zuruecknehmen(i);
            END;
            {
            ** Die Chance, dass ein nachfolgender Zug noch besser ist,
            ** ist gering. Es wird mit einem Null-Fenster weiter
            ** gerechnet, da dadurch die Suche beschleunigt wird.
            }
            alpha:=wert;
            beta:=alpha+1;
            kopiere_hvar(i);
            if not in_Vorschlag then begin
              if english then hilf:=vokabeln[1] else hilf:=vokabeln[0];
               sllines[0]:=hilf+': '+IntToStr(distanz){+' Alpha-Beta Fenster=['+IntToStr(alpha)+','
              +IntToStr(beta)+']'};
              zeige_hvar;
              str(wert/100:0:2, hilf);
              if english then sllines[0]:=slLines[0]+'  '+vokabeln[3]+': '+hilf
              else sllines[0]:=slLines[0]+'  '+vokabeln[2]+': '+hilf;
              if assigned(OnTrace) then OnTrace(self, slLines.Text);
            end;
         END
         else BEGIN { Bereits bisher bester Zug auf Suchtiefe berechnet }
            if(wert>alpha) then BEGIN
               { Neuer bester Zug gefunden. Man weiss im Moment nur,
               ** dass er besser ist. Exakter Wert muss noch einmal mit
               ** offenen Fenster berechnet werden.
               }
               besterwert:=alpha;
               alpha:=wert;
               beta:=MATTWERT;
               zug_ausfuehren(i);
               wert:=-alpha_beta(-beta,-alpha,distanz-1);
               zug_zuruecknehmen(i);
               { Ist er auch mit offenen Fenster besser?.               }
               { Bei reiner Anwendung von alpha-beta muss der Zug       }
               { auch bei offenen Fenster immer besser sein. Da bei     }
               { bei den Erweiterungen und in der Selektivitaet das     }
               { Fenster mit ein geht, kann es in unserem Fall auch     }
               { anders kommen.                                         }
               if(wert>besterwert) then BEGIN
                  alpha:=wert;
                  beta:=alpha+1;
                  kopiere_hvar(i);
                  zeige_hvar;
                  if assigned(OnTrace) then OnTrace(self, slLines.Text);
                  { Plaziere besten Zug an die erste Stelle der Zugliste. }
                  { Schiebe die anderen Zuege um eine Position nach oben  }
                  tmp:=ZugStapel^[i];
                  j:=i;
                  while j>0 do BEGIN
                     ZugStapel^[j]:=ZugStapel^[j-1];
                     dec(j);
                  END;
                  ZugStapel^[0]:=tmp;
               END;
            END;
         END;
         inc(i);
      END;
      inc(distanz);
   END;
   abgebrochen:
   stop:=true;
   if(alpha> -(MATTWERT-1)) then BEGIN
      {zug_ausfuehren(0); der errechnete Zug braucht nicht ausgefhrt zu werden - damit bei Vorschlagberechnung der passende
      Zug von richtiger Stllg. auf Legal. gepr. werden kann}
      if(alpha>=MATTWERT-10) then BEGIN
         if not in_Vorschlag then begin
           sllines[0]:=sllines[0]+'  Matt in '+IntToStr(trunc((MATTWERT-2-alpha)/2));
           if trunc((MATTWERT-2-alpha)/2)=0 then begin
             status:=mit_Matt;sllines[0]:='Matt';
           end;
         end;
      END
      else BEGIN
         if(alpha<= -MATTWERT+10) then BEGIN
            if not in_Vorschlag then begin
              sllines[0]:=sllines[0]+'  Matt in '+IntToStr(trunc((alpha + MATTWERT -1)/2));
              if trunc((MATTWERT-2-alpha)/2)=0 then begin
                status:=mit_Matt;sllines[0]:='Matt';
              end;
            end;
         END;
      END;
      if assigned(OnTrace) then OnTrace(self, slLines.Text);
   END
   else BEGIN
      if(schach=WAHR) then BEGIN
         if not in_Vorschlag then sllines[1]:='MATT!';
         status:=Matt;
      END
      else BEGIN
         if not in_Vorschlag then sllines[1]:='PATT!';
         status:=Patt;
      END;
      if assigned(OnTrace) then OnTrace(self, slLines.Text);
   END;
END;


FUNCTION
TMiniMax.zug_eingabe(von,nach, UFigur : integer; FiRochade, NoMoveGen : Boolean): integer;
var
   i : integer;
   tmp : integer;
BEGIN
   if not NoMoveGen then generiere_zuege(1); {wird nur bei Zugvorschlag nicht gemacht}
   i:=StapelGrenze[Tiefe];
   while i<StapelGrenze[Tiefe+1] do BEGIN
      if((ZugStapel^[i].von=von)AND(ZugStapel^[i].nach=nach)) and
      (
        (not FiRochade) and
        ((ZugStapel^[i].Rochade_nr <> KURZEFISCHERROCHADE) and (ZugStapel^[i].Rochade_Nr <> LANGEFISCHERROCHADE))
        or
        (
          (FiRochade) and
          ((ZugStapel^[i].Rochade_nr = KURZEFISCHERROCHADE) or (ZugStapel^[i].Rochade_Nr = LANGEFISCHERROCHADE))
        )
      )then BEGIN
         if(ZugStapel^[i].umwandl_fig<>LEER) then BEGIN { Umwandlungen werden }
            if abs(UFigur)=W_S then BEGIN                 { in der Reihenfolge  }
               i:=i+1;                         { Dame, Springer, Laeufer }
            END                                 { Turm erzeugt }
            else if abs(Ufigur)=W_L then BEGIN
               i:=i+2;
            END
            else if abs(UFigur)=W_T then BEGIN
               i:=i+3;
            END;
         END;
         init_spielbaum;
         tmp:=LetzterZug;    { Bisher letzten Zug zwischenspeichern. }
         zug_ausfuehren(i); { Achtung: zug_ausfuehren veraendert bereits
                            ** die Farbe. Naechste Farbabfrage muss dies
                            ** kompensieren.
                            }
         if(Farbe=SCHWARZ) then BEGIN
            if(attackiert_feld(wKoenig,SCHWARZ)=WAHR) then BEGIN
               status:=illegaler_Zug;
               zug_zuruecknehmen(i);
               LetzterZug:=tmp;  { Kein neuer Zug gemacht. Bisher letzten }
               zug_eingabe:= FALSCH;   { Zug wieder herstellen.                 }
               exit;
            END;
         END
         else if(attackiert_feld(sKoenig,WEISS)=WAHR) then BEGIN
            status:=illegaler_Zug;
            zug_zuruecknehmen(i);
            LetzterZug:=tmp;
            zug_eingabe:= FALSCH;
            exit;
         END;
         zug_zuruecknehmen(i);  {in Ur-Minimax nicht ausgefhrt - da Stellung immer neu bergeben wird kann das hier geschehen
         - fr Vorschlag wichtig}
         zug_eingabe:= WAHR;
         exit;
      END;
      inc(i);
   END;
   zug_eingabe:= FALSCH;  { Eingegebener Zug wurde in Zugliste nicht gefunden }
END;

constructor TMiniMax.Create(AComponent : TComponent);
begin
  inherited create(AComponent);
  slLines := TStringlist.Create;
  sllines.add('');
  sllines.add('');
  initialisiere;
  stop:=true;
  Vorschlagtiefe:=2;
  Height := 25;
  Width := 25;
end;

destructor TMiniMax.Destroy;
begin
  slLines.Free;
  inherited Destroy;
end;

Procedure TMiniMax.PStop;
var MM1Zug : PZug; {da MMZug typisierter Pointer ist}
begin
  zug_ausfuehren(0);   {damit bewegt[] gesetzt wird wegen Aktualisierung d. Rochadem.}
  if -MMZug.amZug = SCHWARZ then begin
    if attackiert_feld(skoenig, WEISS)=WAHR then if status = 0 then status := mit_schach;
  end
  else begin
    if attackiert_feld(wkoenig, SCHWARZ)=WAHR then if status = 0 then status := mit_schach;
  end;
  with MMZug do begin
    von:=zugstapel^[0].von;
    nach:=zugstapel^[0].nach;
    if (epFeld<>ILLEGAL) and (zugstapel^[0].geschl_Fig<>LEER) then begin  {dann ist es ep-Schlag}
      epSchlag:=zugstapel^[0].ep_Feld; {in dem Fall enthlt epFeld das des zu schlagenden Bauern}
      epFeld:=LEER;                   {und es kann kein Doppelschritt gewesen sein}
    end
    else begin
      epSchlag:=LEER;
      epFeld:=zugstapel^[0].ep_Feld;  {das Feld hinter dem Doppelschrittbauern oder 0}
    end;
    geschlFigur:=zugstapel^[0].geschl_Fig*-amZug;
    UFigur:=Zugstapel^[0].Umwandl_fig*amZug;
    rochadeArt:=ZugStapel^[0].Rochade_Nr;
  end;
  if assigned(FRC) then with FRC^ do begin
    if bewegt[FischerWhiteKing]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WkR, WlR];
    if bewegt[FischerLeftWhiteRook]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WlR];
    if bewegt[FischerRightWhiteRook]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WkR];
    if bewegt[FischerBlackKing]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SkR, SlR];
    if bewegt[FischerLeftBlackRook]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SlR];
    if bewegt[FischerRightBlackRook]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SkR];
  end
  else begin
    if bewegt[e1]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WkR, WlR];
    if bewegt[a1]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WlR];
    if bewegt[h1]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[WkR];
    if bewegt[e8]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SkR, SlR];
    if bewegt[a8]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SlR];
    if bewegt[h8]<>0 then MMZug.Rochademoeglichkeiten:=MMZug.Rochademoeglichkeiten-[SkR];
  end;
  MMZug.amZug:=-MMZug.amZug;
  MM1Zug:=@MMZug;
  Dispose(ZugStapel);
  slLines.Text := '';
  sllines.add('');
  sllines.add('');
  stop:=true;
  if not FBreak then begin
    if Assigned(FOnStop) then FOnStop(Self, MM1Zug, status);
  end  
  else FBreak := false;
end;

procedure TMiniMax.StellungsDatenEintragen;
Var rwrook, lwrook, rbrook, lbrook: shortint;
begin
  stellungs_eingabe;
  if assigned(FRC) then with FRC^ do begin
    rwrook := FischerRightWhiteRook;
    lwrook := FischerLeftWhiteRook;
    rbrook := FischerRightBlackRook;
    lbrook := FischerLeftBlackRook;
  end
  else begin
    rwrook := h1;
    lwrook := a1;
    rbrook := h8;
    lbrook := a8;
  end;
  if wkr in MMZug.Rochademoeglichkeiten  then bewegt[rwrook]:=0;
  if wlr in MMZug.Rochademoeglichkeiten  then bewegt[lwrook]:=0;
  if skr in MMZug.Rochademoeglichkeiten  then bewegt[rbrook]:=0;
  if slr in MMZug.Rochademoeglichkeiten  then bewegt[lbrook]:=0;
  if (wkr in MMZug.Rochademoeglichkeiten) and (wlr in MMZug.Rochademoeglichkeiten) {Rochaderecht verloren}
    then rochade[weiss+1]:=FALSCH else rochade[weiss+1]:=WAHR;
  if (skr in MMZug.Rochademoeglichkeiten) and (slr in MMZug.Rochademoeglichkeiten) {Rochaderecht verloren}
    then rochade[schwarz+1]:=FALSCH else rochade[schwarz+1]:=WAHR;
  ep_Feld[0]:=MMZug.epFeld;
end;

procedure
  TMiniMax.Start(stellung : TSchachbrett;Zug : TZug; t:byte);
begin
 if stop then begin  {damit Schleife nicht 2 Mal gestartet wird}
   New(ZugStapel);  {wird in PStop gelscht}
   Status:=0;     {zeigt bei Stopereignis z.B. Matt, Illegaler_Zug usw. an}
   Solltiefe:=t; {30-10 wegen Distanzerweiterung}
   MMBrett := Stellung;
   MMZug := Zug;
   farbe:=MMZug.amZug; {Vorzeichen sind bei "Start" u. "IstZuglegal" umgedreht}
   StellungsDatenEintragen;
   if MMZug.amZug = SCHWARZ then begin
    if attackiert_feld(skoenig, WEISS)=WAHR then if assigned(OnSchach) then OnSchach(self);
   end
   else begin
     if attackiert_feld(wkoenig, SCHWARZ)=WAHR then if assigned(OnSchach) then OnSchach(self);
   end;
   Stop:=false;
   FBreak := false;
   computer_zug;
   akzeptabel;  {da sich letzte Stellungen gemerkt werden, kommt es bei mehrmaligen Rcksetzen eines Mattzuges vor, da beim 3.Mal ein anderer ausgefhrt wird!!!}
   {soll Bug bei Patt/Matterkennung bei Zugforderung ausbgeln... Art und Weise nicht gerade optimal...}
   if (stop) and ((status=patt) or (status=matt)) then begin
     stop:=false;
     Solltiefe:=1;
     Status:=0;
     computer_zug;
   end;
   PStop;   {ruft Ereignis OnStop auf}
 end;
end;

function TMiniMax.IstZugLegal(stellung : TSchachbrett;Zug : TZug) : Boolean;
Var FiRochade : Boolean;
begin
  New(ZugStapel);
  MMBrett := Stellung;
  MMZug := Zug;
  farbe:=MMZug.amZug;
  StellungsDatenEintragen;
  if (assigned(FRC)) and (MMZug.RochadeArt <> KEINEROCHADE) then FiRochade := true
  else FiRochade := false;
  if zug_eingabe(MMZug.von, MMZug.nach, 0, FiRochade, false)=FALSCH {bei Zugprfung ist UFigur egal}
  then IstZugLegal:=false else IstZugLegal:=true;
  Dispose(ZugStapel);
end;

function
  TMiniMax.Vorschlag(stellung : TSchachbrett;Zug : TZug; feld : byte) : byte;
Var i : word;
    ZV : Byte;
begin
 Result := 0;
 if stop then begin  {damit Schleife nicht 2 Mal gestartet wird}
   New(ZugStapel);
   Status:=0;     {zeigt bei Stopereignis z.B. Matt, Illegaler_Zug usw. an}
   Solltiefe:=Vorschlagtiefe;
   MMBrett := Stellung;
   MMZug := Zug;
   StellungsDatenEintragen;
   farbe:=MMZug.amZug; {Vorzeichen sind bei "Start" u. "IstZuglegal" umgedreht}
   Stop:=false;
   FBreak := false;
   in_vorschlag:=true;
   computer_zug;
   in_Vorschlag:=false;
   i:=0;
   ZV:=0;
   while (i<=StapelGrenze[1]) and (ZV=0) do BEGIN
     {keine Figur der Seite, die amZug ist oder Leer?}
     if (MMBrett[feld]=LEER) or (not(MMBrett[feld]=MMZug.amZug*abs(MMBrett[feld])))then begin
       if ZugStapel^[i].nach=Feld then ZV:=ZugStapel^[i].von;
     end
     else if ZugStapel^[i].von=Feld then ZV:=ZugStapel^[i].nach;
     if zug_eingabe(ZugStapel^[i].von, ZugStapel^[i].nach, 0, false, true)=FALSCH then ZV:=0; {im Zugstapel sind auch illegale Zge - deshalb Prfung
     (Stllg. wurde nicht verndert}
     inc(i);
   end;
   Dispose(ZugStapel);
   vorschlag:=ZV;
 end;
end;

{!!!Routine soll bei bevorstehender Zugwiederholung in besserer Stellung den zweitbesten Zug
    suchen. Dabei soll dieser keinen negativen Wert besitzen. Im Zugstapel ist das jedoch
    bei Tiefe 1 ungengend vermerkt, so da ein schlechterer Zug als Zugwiederholung vorkommen
    knnte. Die Situation allgemein sollte jedoch selten auftreten.}

procedure TMiniMax.akzeptabel;
var i, w : integer;
    aw : real;
    gleich, moegl_Zug : boolean;
    zuege : array[0..79] of zug_t; {mehr als 80 Zugmglichkeiten sollte es nicht geben}
begin
  aw:=aktuellerWert;
  if aw<0.5 then exit;
  aw:=aw*100;
  gleich:=true;
  for i:=0 to 119 do if MMBrett[i]<>rpt3S[2][i] then gleich:=false;
  with ZugStapel^[0] do begin
    if (von <> rpt3Z[2].von) or (nach <> rpt3Z[2].nach) or (geschl_fig <> rpt3Z[2].geschl_Fig) or
    (umwandl_fig <> rpt3Z[2].umwandl_fig) or (rochade_nr <> rpt3Z[2].rochade_nr) or
    (ep_feld <> rpt3Z[2].ep_feld) then gleich:=false;
  end;
  if gleich then begin  {dann sind Stellungen und Zge gleich --> was anderes probieren}
    i:=0;
    repeat              {weil Zugstapel bei Zug_eingabe verndert wird}
      zuege[i]:=ZugStapel^[i];
      inc(i);
    until (i>StapelGrenze[1]) or (i>79);
    i:=1;
    moegl_Zug:=false;
    case abs(MMBrett[zuege[0].von]) of  {Wert der Figur, die eigentlich ziehen soll}
      wB  : w:=MAT_B;
      wT  : w:=MAT_T;
      wL  : w:=MAT_L;
      wS  : w:=MAT_S;
      wD  : w:=MAT_D;
      else  w:=1000;
    end;
    while (i<=StapelGrenze[1]) and (moegl_Zug=false) do BEGIN
      if (zug_eingabe(zuege[i].von, zuege[i].nach, 0, false, false)=WAHR){im Zugstapel sind auch illegale Zge - deshalb Prfung}
      and (zuege[i].wert-w+aw>0) then moegl_Zug:=true;  {nur Zge beachten, die noch zu besserer Stellung fhren}
      inc(i);
    end;
    dec(i);
    if (moegl_zug) and (i>0) and (i<=79) then zugstapel^[0]:=zuege[i]
    else zugstapel^[0]:=zuege[0];
  end;
  rpt3S[2]:=rpt3S[1];
  rpt3S[1]:=MMBrett;
  rpt3Z[2]:=rpt3Z[1];
  rpt3Z[1]:=ZugStapel^[0];
end;

procedure TMiniMax.BreakLoop;
begin
  FBreak := true;
  FStop := true;
end;

function TMinimax.aktuellerWert : real;
var i : byte;
    hilf : string;
    wert : real;
    fehler : integer;
begin
  aktuellerWert:=KEINWERT;
  hilf:='';
  i:=length(sllines[0]);
  while (i>0) and (sllines[0][i]<>':') do begin
    hilf:=sllines[0][i]+hilf;
    dec(I);
  end;
  val(hilf, wert, fehler);
  if fehler=0 then aktuellerWert:=Wert;
end;

initialization
  aBitmap := nil;
  Loaded := False;
finalization
  aBitmap.Free;
end.

